# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
from abc import ABCMeta, abstractmethod
from datetime import date, datetime
from typing import TYPE_CHECKING

from typing_extensions import Concatenate

try:
    from typing import ParamSpec
except Exception:
    from typing_extensions import ParamSpec  # type: ignore

import pandas as pd

from pylegend._typing import (
    PyLegendSequence,
    PyLegendTypeVar,
    PyLegendType,
    PyLegendList,
    PyLegendTuple,
    PyLegendSet,
    PyLegendOptional,
    PyLegendCallable,
    PyLegendUnion,
    PyLegendDict
)
from pylegend.core.database.sql_to_string import (
    SqlToStringConfig,
    SqlToStringFormat
)
from pylegend.core.language import (
    PyLegendPrimitive,
    PyLegendInteger,
    PyLegendBoolean,
)
from pylegend.core.language.pandas_api.pandas_api_aggregate_specification import PyLegendAggInput
from pylegend.core.language.pandas_api.pandas_api_tds_row import PandasApiTdsRow
from pylegend.core.language.shared.primitives.primitive import PyLegendPrimitiveOrPythonPrimitive
from pylegend.core.language.shared.tds_row import AbstractTdsRow
from pylegend.core.sql.metamodel import QuerySpecification
from pylegend.core.tds.abstract.frames.base_tds_frame import BaseTdsFrame
from pylegend.core.tds.pandas_api.frames.pandas_api_tds_frame import PandasApiTdsFrame
from pylegend.core.tds.result_handler import (
    ResultHandler,
    ToStringResultHandler,
)
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToPureConfig
from pylegend.core.tds.tds_frame import FrameToSqlConfig, PyLegendTdsFrame
from pylegend.extensions.tds.result_handler import (
    ToPandasDfResultHandler,
    PandasDfReadConfig,
)

if TYPE_CHECKING:
    from pylegend.core.language.pandas_api.pandas_api_series import Series
    from pylegend.core.tds.pandas_api.frames.pandas_api_groupby_tds_frame import PandasApiGroupbyTdsFrame

__all__: PyLegendSequence[str] = [
    "PandasApiBaseTdsFrame"
]

R = PyLegendTypeVar('R')
P = ParamSpec("P")


class PandasApiBaseTdsFrame(PandasApiTdsFrame, BaseTdsFrame, metaclass=ABCMeta):
    __columns: PyLegendSequence[TdsColumn]

    def __init__(self, columns: PyLegendSequence[TdsColumn]) -> None:
        col_names = [c.get_name() for c in columns]
        if len(col_names) != len(set(col_names)):
            cols = "[" + ", ".join([str(c) for c in columns]) + "]"
            raise ValueError(f"TdsFrame cannot have duplicated column names. Passed columns: {cols}")
        self.__columns = [c.copy() for c in columns]
        self._transformed_frame = None

    def columns(self) -> PyLegendSequence[TdsColumn]:
        if self._transformed_frame is None:
            return [c.copy() for c in self.__columns]
        return self._transformed_frame.columns()

    def __getitem__(
            self,
            key: PyLegendUnion[str, PyLegendList[str], PyLegendBoolean]
    ) -> PyLegendUnion["PandasApiTdsFrame", "Series"]:
        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import \
            PandasApiAppliedFunctionTdsFrame
        from pylegend.core.tds.pandas_api.frames.functions.filtering import \
            PandasApiFilteringFunction
        from pylegend.core.language.shared.primitives.boolean import PyLegendBoolean

        if isinstance(key, PyLegendBoolean):
            return PandasApiAppliedFunctionTdsFrame(
                PandasApiFilteringFunction(self, filter_expr=key)
            )

        elif isinstance(key, str):
            for col in self.__columns:
                if col.get_name() == key:
                    col_type = col.get_type()
                    if col_type == "Boolean":
                        from pylegend.core.language.pandas_api.pandas_api_series import \
                            BooleanSeries  # pragma: no cover
                        return BooleanSeries(self, key)  # pragma: no cover (Boolean column not supported in PURE)
                    elif col_type == "String":
                        from pylegend.core.language.pandas_api.pandas_api_series import StringSeries
                        return StringSeries(self, key)
                    elif col_type == "Integer":
                        from pylegend.core.language.pandas_api.pandas_api_series import IntegerSeries
                        return IntegerSeries(self, key)
                    elif col_type == "Float":
                        from pylegend.core.language.pandas_api.pandas_api_series import FloatSeries
                        return FloatSeries(self, key)
                    elif col_type == "Date":
                        from pylegend.core.language.pandas_api.pandas_api_series import DateSeries
                        return DateSeries(self, key)
                    elif col_type == "DateTime":
                        from pylegend.core.language.pandas_api.pandas_api_series import DateTimeSeries
                        return DateTimeSeries(self, key)
                    elif col_type == "StrictDate":
                        from pylegend.core.language.pandas_api.pandas_api_series import StrictDateSeries
                        return StrictDateSeries(self, key)
                    else:
                        raise ValueError(f"Unsupported column type '{col_type}' for column '{key}'")  # pragma: no cover
            raise KeyError(f"['{key}'] not in index")

        elif isinstance(key, list):
            valid_col_names = {col.get_name() for col in self.__columns}
            invalid_cols = [k for k in key if k not in valid_col_names]
            if invalid_cols:
                raise KeyError(f"{invalid_cols} not in index")
            return self.filter(items=key)
        else:
            raise TypeError(f"Invalid key type: {type(key)}. Expected str, list, or boolean expression")

    def __setitem__(self, key: str, value: PyLegendUnion["Series", PyLegendPrimitiveOrPythonPrimitive]) -> None:
        """
        Pandas-like column assignment with replace semantics:
        - If column exists, drop it first.
        - Then assign the new value (Series or constant).
        """
        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
            PandasApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.pandas_api.frames.functions.assign_function import AssignFunction
        from pylegend.core.language.pandas_api.pandas_api_series import Series

        # Type Check
        if not isinstance(key, str):
            raise TypeError(f"Column name must be a string, got: {type(key)}")

        # Reject cross-frame assignment
        if isinstance(value, Series):
            origin = value.get_base_frame()
            if origin is not None and origin is not self:
                raise ValueError("Assignment from a different frame is not allowed")

        # Normalize the assignment value
        col_def = {}
        if callable(value):
            col_def[key] = value
        else:
            col_def[key] = lambda row: value

        working_frame = copy.deepcopy(self)
        assign_applied = PandasApiAppliedFunctionTdsFrame(AssignFunction(working_frame, col_definitions=col_def))

        self._transformed_frame = assign_applied  # type: ignore
        self.__columns = assign_applied.columns()

    def assign(
            self,
            **kwargs: PyLegendCallable[
                [PandasApiTdsRow],
                PyLegendUnion[int, float, bool, str, date, datetime, PyLegendPrimitive]
            ],
    ) -> "PandasApiTdsFrame":
        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
            PandasApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.pandas_api.frames.functions.assign_function import AssignFunction
        return PandasApiAppliedFunctionTdsFrame(AssignFunction(self, col_definitions=kwargs))

    def filter(
            self,
            items: PyLegendOptional[PyLegendList[str]] = None,
            like: PyLegendOptional[str] = None,
            regex: PyLegendOptional[str] = None,
            axis: PyLegendOptional[PyLegendUnion[str, int, PyLegendInteger]] = None
    ) -> "PandasApiTdsFrame":
        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
            PandasApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.pandas_api.frames.functions.filter import PandasApiFilterFunction
        return PandasApiAppliedFunctionTdsFrame(
            PandasApiFilterFunction(
                self,
                items=items,
                like=like,
                regex=regex,
                axis=axis
            )
        )

    def sort_values(
            self,
            by: PyLegendUnion[str, PyLegendList[str]],
            axis: PyLegendUnion[str, int] = 0,
            ascending: PyLegendUnion[bool, PyLegendList[bool]] = True,
            inplace: bool = False,
            kind: PyLegendOptional[str] = None,
            na_position: str = 'last',
            ignore_index: bool = True,
            key: PyLegendOptional[PyLegendCallable[[AbstractTdsRow], AbstractTdsRow]] = None
    ) -> "PandasApiTdsFrame":
        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
            PandasApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.pandas_api.frames.functions.sort_values_function import SortValuesFunction
        return PandasApiAppliedFunctionTdsFrame(SortValuesFunction(
            base_frame=self,
            by=by,
            axis=axis,
            ascending=ascending,
            inplace=inplace,
            kind=kind,
            na_position=na_position,
            ignore_index=ignore_index,
            key=key
        ))

    def truncate(
            self,
            before: PyLegendUnion[date, str, int, None] = None,
            after: PyLegendUnion[date, str, int, None] = None,
            axis: PyLegendUnion[str, int] = 0,
            copy: bool = True
    ) -> "PandasApiTdsFrame":
        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
            PandasApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.pandas_api.frames.functions.truncate_function import TruncateFunction
        return PandasApiAppliedFunctionTdsFrame(TruncateFunction(
            base_frame=self,
            before=before,
            after=after,
            axis=axis,
            copy=copy
        ))

    def drop(
            self,
            labels: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str], PyLegendSet[str]]] = None,
            axis: PyLegendUnion[str, int, PyLegendInteger] = 1,
            index: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str], PyLegendSet[str]]] = None,
            columns: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str], PyLegendSet[str]]] = None,
            level: PyLegendOptional[PyLegendUnion[int, PyLegendInteger, str]] = None,
            inplace: PyLegendUnion[bool, PyLegendBoolean] = False,
            errors: str = "raise",
    ) -> "PandasApiTdsFrame":
        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import \
            PandasApiAppliedFunctionTdsFrame
        from pylegend.core.tds.pandas_api.frames.functions.drop import PandasApiDropFunction

        return PandasApiAppliedFunctionTdsFrame(
            PandasApiDropFunction(
                base_frame=self,
                labels=labels,
                axis=axis,
                index=index,
                columns=columns,
                level=level,
                inplace=inplace,
                errors=errors
            )
        )

    def aggregate(
            self,
            func: PyLegendAggInput,
            axis: PyLegendUnion[int, str] = 0,
            *args: PyLegendPrimitiveOrPythonPrimitive,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
            PandasApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.pandas_api.frames.functions.aggregate_function import AggregateFunction
        return PandasApiAppliedFunctionTdsFrame(AggregateFunction(
            self,
            func,
            axis,
            *args,
            **kwargs
        ))

    def agg(
            self,
            func: PyLegendAggInput,
            axis: PyLegendUnion[int, str] = 0,
            *args: PyLegendPrimitiveOrPythonPrimitive,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
            PandasApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.pandas_api.frames.functions.aggregate_function import AggregateFunction
        return PandasApiAppliedFunctionTdsFrame(AggregateFunction(
            self,
            func,
            axis,
            *args,
            **kwargs
        ))

    def sum(
            self,
            axis: PyLegendUnion[int, str] = 0,
            skipna: bool = True,
            numeric_only: bool = False,
            min_count: int = 0,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        if axis not in [0, "index"]:
            raise NotImplementedError(f"The 'axis' parameter must be 0 or 'index' in sum function, but got: {axis}")
        if skipna is not True:
            raise NotImplementedError("skipna=False is not currently supported in sum function. "
                                      "SQL aggregation ignores nulls by default.")
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in sum function.")
        if min_count != 0:
            raise NotImplementedError(f"min_count must be 0 in sum function, but got: {min_count}")
        if len(kwargs) > 0:
            raise NotImplementedError(
                f"Additional keyword arguments not supported in sum function: {list(kwargs.keys())}")
        return self.aggregate("sum", 0)

    def mean(
            self,
            axis: PyLegendUnion[int, str] = 0,
            skipna: bool = True,
            numeric_only: bool = False,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        if axis not in [0, "index"]:
            raise NotImplementedError(f"The 'axis' parameter must be 0 or 'index' in mean function, but got: {axis}")
        if skipna is not True:
            raise NotImplementedError("skipna=False is not currently supported in mean function.")
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in mean function.")
        if len(kwargs) > 0:
            raise NotImplementedError(
                f"Additional keyword arguments not supported in mean function: {list(kwargs.keys())}")
        return self.aggregate("mean", 0)

    def min(
            self,
            axis: PyLegendUnion[int, str] = 0,
            skipna: bool = True,
            numeric_only: bool = False,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        if axis not in [0, "index"]:
            raise NotImplementedError(f"The 'axis' parameter must be 0 or 'index' in min function, but got: {axis}")
        if skipna is not True:
            raise NotImplementedError("skipna=False is not currently supported in min function.")
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in min function.")
        if len(kwargs) > 0:
            raise NotImplementedError(
                f"Additional keyword arguments not supported in min function: {list(kwargs.keys())}")
        return self.aggregate("min", 0)

    def max(
            self,
            axis: PyLegendUnion[int, str] = 0,
            skipna: bool = True,
            numeric_only: bool = False,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        if axis not in [0, "index"]:
            raise NotImplementedError(f"The 'axis' parameter must be 0 or 'index' in max function, but got: {axis}")
        if skipna is not True:
            raise NotImplementedError("skipna=False is not currently supported in max function.")
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in max function.")
        if len(kwargs) > 0:
            raise NotImplementedError(
                f"Additional keyword arguments not supported in max function: {list(kwargs.keys())}")
        return self.aggregate("max", 0)

    def std(
            self,
            axis: PyLegendUnion[int, str] = 0,
            skipna: bool = True,
            ddof: int = 1,
            numeric_only: bool = False,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        if axis not in [0, "index"]:
            raise NotImplementedError(f"The 'axis' parameter must be 0 or 'index' in std function, but got: {axis}")
        if skipna is not True:
            raise NotImplementedError("skipna=False is not currently supported in std function.")
        if ddof != 1:
            raise NotImplementedError(
                f"Only ddof=1 (Sample Standard Deviation) is supported in std function, but got: {ddof}")
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in std function.")
        if len(kwargs) > 0:
            raise NotImplementedError(
                f"Additional keyword arguments not supported in std function: {list(kwargs.keys())}")
        return self.aggregate("std", 0)

    def var(
            self,
            axis: PyLegendUnion[int, str] = 0,
            skipna: bool = True,
            ddof: int = 1,
            numeric_only: bool = False,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        if axis not in [0, "index"]:
            raise NotImplementedError(f"The 'axis' parameter must be 0 or 'index' in var function, but got: {axis}")
        if skipna is not True:
            raise NotImplementedError("skipna=False is not currently supported in var function.")
        if ddof != 1:
            raise NotImplementedError(f"Only ddof=1 (Sample Variance) is supported in var function, but got: {ddof}")
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in var function.")
        if len(kwargs) > 0:
            raise NotImplementedError(
                f"Additional keyword arguments not supported in var function: {list(kwargs.keys())}")
        return self.aggregate("var", 0)

    def count(
            self,
            axis: PyLegendUnion[int, str] = 0,
            numeric_only: bool = False,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        if axis not in [0, "index"]:
            raise NotImplementedError(f"The 'axis' parameter must be 0 or 'index' in count function, but got: {axis}")
        if numeric_only is not False:
            raise NotImplementedError("numeric_only=True is not currently supported in count function.")
        if len(kwargs) > 0:
            raise NotImplementedError(
                f"Additional keyword arguments not supported in count function: {list(kwargs.keys())}")
        return self.aggregate("count", 0)

    def groupby(
            self,
            by: PyLegendUnion[str, PyLegendList[str]],
            level: PyLegendOptional[PyLegendUnion[str, int, PyLegendList[str]]] = None,
            as_index: bool = False,
            sort: bool = True,
            group_keys: bool = False,
            observed: bool = False,
            dropna: bool = False,
    ) -> "PandasApiGroupbyTdsFrame":
        from pylegend.core.tds.pandas_api.frames.pandas_api_groupby_tds_frame import (
            PandasApiGroupbyTdsFrame
        )
        return PandasApiGroupbyTdsFrame(
            base_frame=self,
            by=by,
            level=level,
            as_index=as_index,
            sort=sort,
            group_keys=group_keys,
            observed=observed,
            dropna=dropna
        )

    def merge(
            self,
            other: "PandasApiTdsFrame",
            how: PyLegendOptional[str] = "inner",
            on: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str]]] = None,
            left_on: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str]]] = None,
            right_on: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str]]] = None,
            left_index: PyLegendOptional[bool] = False,
            right_index: PyLegendOptional[bool] = False,
            sort: PyLegendOptional[bool] = False,
            suffixes: PyLegendOptional[
                PyLegendUnion[
                    PyLegendTuple[PyLegendUnion[str, None], PyLegendUnion[str, None]],
                    PyLegendList[PyLegendUnion[str, None]],
                ]
            ] = ("_x", "_y"),
            indicator: PyLegendOptional[PyLegendUnion[bool, str]] = False,
            validate: PyLegendOptional[str] = None
    ) -> "PandasApiTdsFrame":
        """
        Pandas-like merge:
        - Mutually exclusive: `on` vs (`left_on`, `right_on`)
        - If no keys provided, infer intersection of column names
        - `how`: inner | left | right | outer (outer mapped to full)
        - `suffixes`: applied to overlapping non-key columns
        """
        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
            PandasApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.pandas_api.frames.functions.merge import (
            PandasApiMergeFunction
        )
        merge_fn = PandasApiMergeFunction(
            self,
            other,  # type: ignore
            how=how,
            on=on,
            left_on=left_on,
            right_on=right_on,
            left_index=left_index,
            right_index=right_index,
            sort=sort,
            suffixes=suffixes,
            indicator=indicator,
            validate=validate
        )
        merged = PandasApiAppliedFunctionTdsFrame(merge_fn)

        if sort:
            return merged.sort_values(
                by=merge_fn.get_sort_keys(),
                axis=0,
                ascending=True,
                inplace=False,
                kind=None,
                na_position="last",
                ignore_index=True,
                key=None
            )
        else:
            return merged

    def join(
            self,
            other: "PandasApiTdsFrame",
            on: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str]]] = None,
            how: PyLegendOptional[str] = "left",
            lsuffix: str = "",
            rsuffix: str = "",
            sort: PyLegendOptional[bool] = False,
            validate: PyLegendOptional[str] = None
    ) -> "PandasApiTdsFrame":
        """
        Pandas-like join delegating to merge. No index support, only column-on-column via `on`.
        """
        return self.merge(
            other=other,
            how=how,
            on=on,
            sort=sort,
            suffixes=[lsuffix, rsuffix],
            validate=validate
        )

    def rename(
            self,
            mapper: PyLegendOptional[PyLegendUnion[PyLegendDict[str, str], PyLegendCallable[[str], str]]] = None,
            index: PyLegendOptional[PyLegendUnion[PyLegendDict[str, str], PyLegendCallable[[str], str]]] = None,
            columns: PyLegendOptional[PyLegendUnion[PyLegendDict[str, str], PyLegendCallable[[str], str]]] = None,
            axis: PyLegendUnion[str, int] = 1,
            inplace: PyLegendUnion[bool] = False,
            copy: PyLegendUnion[bool] = True,
            level: PyLegendOptional[PyLegendUnion[int, str]] = None,
            errors: str = "ignore",
    ) -> "PandasApiTdsFrame":
        """
        Pandas-like rename:
        - Supports mapping via `mapper` or explicit `index`/`columns`
        - Only column renames are applied when `axis` is 1
        - `errors`: ignore | raise
        """

        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
            PandasApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.pandas_api.frames.functions.rename import (
            PandasApiRenameFunction
        )
        return PandasApiAppliedFunctionTdsFrame(
            PandasApiRenameFunction(
                base_frame=self,
                mapper=mapper,
                axis=axis,
                index=index,
                columns=columns,
                copy=copy,
                inplace=inplace,
                level=level,
                errors=errors
            )
        )

    def apply(
            self,
            func: PyLegendUnion[
                PyLegendCallable[Concatenate["Series", P], PyLegendPrimitiveOrPythonPrimitive],
                str
            ],
            axis: PyLegendUnion[int, str] = 0,
            raw: bool = False,
            result_type: PyLegendOptional[str] = None,
            args: PyLegendTuple[PyLegendPrimitiveOrPythonPrimitive, ...] = (),
            by_row: PyLegendUnion[bool, str] = "compat",
            engine: str = "python",
            engine_kwargs: PyLegendOptional[PyLegendDict[str, PyLegendPrimitiveOrPythonPrimitive]] = None,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        """
        Pandas-like apply (columns-only):
        - Supports callable func applied to each column (axis=0 or 'index')
        - Internally delegates to assign by constructing lambdas per column
        - Unsupported params raise NotImplementedError
        """

        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
            PandasApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.pandas_api.frames.functions.assign_function import AssignFunction
        from pylegend.core.language.pandas_api.pandas_api_series import Series

        # Validation
        if axis not in (0, "index"):
            raise ValueError("Only column-wise apply is supported. Use axis=0 or 'index'")
        if raw:
            raise NotImplementedError("raw=True is not supported. Use raw=False")
        if result_type is not None:
            raise NotImplementedError("result_type is not supported")
        if by_row not in (False, "compat"):
            raise NotImplementedError("by_row must be False or 'compat'")
        if engine != "python":
            raise NotImplementedError("Only engine='python' is supported")
        if engine_kwargs is not None:
            raise NotImplementedError("engine_kwargs are not supported")
        if isinstance(func, str):
            raise NotImplementedError("String-based apply is not supported")
        if not callable(func):
            raise TypeError("Function must be a callable")

        # Build assign column definitions: apply func to each column Series
        col_definitions = {}
        for c in self.columns():
            col_name = c.get_name()
            series = self[col_name]

            # Compute row callable via func on the Series
            def _row_callable(
                    _row: PandasApiTdsRow,
                    _s: Series = series,  # type: ignore
                    _a: PyLegendTuple[PyLegendPrimitiveOrPythonPrimitive, ...] = args,
                    _k: PyLegendPrimitiveOrPythonPrimitive = kwargs  # type: ignore
            ) -> PyLegendPrimitiveOrPythonPrimitive:
                return func(_s, *_a, **_k)  # type: ignore

            col_definitions[col_name] = _row_callable

        return PandasApiAppliedFunctionTdsFrame(
            AssignFunction(self, col_definitions=col_definitions)  # type: ignore
        )

    def head(self, n: int = 5) -> "PandasApiTdsFrame":
        """
        Return the first `n` rows by calling truncate on rows.
        Negative `n` is not supported.
        """
        if not isinstance(n, int):
            raise TypeError(f"n must be an int, got {type(n)}")
        if n < 0:
            raise NotImplementedError("Negative n is not supported yet in Pandas API head")

        return self.truncate(before=None, after=max(n - 1, -1), axis=0, copy=True)

    @property
    def shape(self) -> PyLegendTuple[int, int]:
        """
        Return a tuple representing the dimensionality of the TdsFrame
        as (number of rows, number of columns).
        """

        col_name = self.columns()[0].get_name()
        newframe = self.aggregate(func={col_name: "count"}, axis=0)

        df = newframe.execute_frame_to_pandas_df()

        total_rows = df.iloc[0, 0]
        total_cols = len(self.columns())

        return (total_rows, total_cols)  # type: ignore

    def dropna(
            self,
            axis: PyLegendUnion[int, str] = 0,
            how: str = "any",
            thresh: PyLegendOptional[int] = None,
            subset: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str]]] = None,
            inplace: bool = False,
            ignore_index: bool = False
    ) -> "PandasApiTdsFrame":
        """
        Remove missing values.

        Parameters
        ----------
        axis : {0 or 'index'}, default 0
            Determine if rows or columns which contain missing values are removed.
            * 0, or 'index' : Drop rows which contain missing values.
            Currently, only `axis=0` is supported.
        how : {'any', 'all'}, default 'any'
            Determine if row is removed from TdsFrame, when we have at least one NA or all NA.
            * 'any' : If any NA values are present, drop that row.
            * 'all' : If all values are NA, drop that row.
        thresh : int, optional
            Not implemented yet.
        subset : list-like, optional
            Labels along other axis to consider, e.g. if you are dropping rows
            these would be a list of columns to include.
        inplace : bool, default False
            Not implemented yet.
        ignore_index : bool, default False
            Not implemented yet.

        Returns
        -------
        PandasApiTdsFrame
            TdsFrame with NA entries dropped.
        """
        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
            PandasApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.pandas_api.frames.functions.dropna import PandasApiDropnaFunction
        return PandasApiAppliedFunctionTdsFrame(
            PandasApiDropnaFunction(
                base_frame=self,
                axis=axis,
                how=how,
                thresh=thresh,
                subset=subset,
                inplace=inplace,
                ignore_index=ignore_index
            )
        )

    def fillna(
            self,
            value: PyLegendUnion[
                int, float, str, bool, date, datetime,
                PyLegendDict[str, PyLegendUnion[int, float, str, bool, date, datetime]]
            ] = None,  # type: ignore
            axis: PyLegendOptional[PyLegendUnion[int, str]] = 0,
            inplace: bool = False,
            limit: PyLegendOptional[int] = None
    ) -> "PandasApiTdsFrame":
        """
        Fill missing values.

        Parameters
        ----------
        base_frame : PandasApiBaseTdsFrame
            The base frame to apply fillna on.
        value : scalar, dict, default None
            Value to use to fill holes (e.g. 0), alternately a dict of values specifying
            which value to use for each column of TdsFrame.
        axis : {0 or 'index'}, default 0
            Axis along which to fill missing values.
            * 0, or 'index' : Fill missing values for each column.
            Currently, only `axis=0` is supported.
        inplace : bool, default False
            Not implemented yet.
        limit : int, optional
            Not implemented yet.

        Returns
        -------
        PandasApiTdsFrame
            TdsFrame with NA entries filled.
        """
        from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
            PandasApiAppliedFunctionTdsFrame
        )
        from pylegend.core.tds.pandas_api.frames.functions.fillna import PandasApiFillnaFunction
        return PandasApiAppliedFunctionTdsFrame(
            PandasApiFillnaFunction(
                base_frame=self,
                value=value,
                axis=axis,
                inplace=inplace,
                limit=limit
            )
        )

    @abstractmethod
    def get_super_type(self) -> PyLegendType[PyLegendTdsFrame]:
        pass  # pragma: no cover

    def to_sql_query_object(self, config: FrameToSqlConfig) -> QuerySpecification:
        if self._transformed_frame is None:
            return self.get_super_type().to_sql_query_object(self, config)  # type: ignore
        else:
            return self._transformed_frame.to_sql_query_object(config)

    def to_pure(self, config: FrameToPureConfig) -> str:
        if self._transformed_frame is None:
            return self.get_super_type().to_pure(self, config)  # type: ignore
        else:
            return self._transformed_frame.to_pure(config)

    def to_pure_query(self, config: FrameToPureConfig = FrameToPureConfig()) -> str:
        return self.to_pure(config)

    @abstractmethod
    def get_all_tds_frames(self) -> PyLegendList["PandasApiBaseTdsFrame"]:
        pass  # pragma: no cover

    def to_sql_query(self, config: FrameToSqlConfig = FrameToSqlConfig()) -> str:
        query = self.to_sql_query_object(config)
        sql_to_string_config = SqlToStringConfig(
            format_=SqlToStringFormat(
                pretty=config.pretty
            )
        )
        return config.sql_to_string_generator().generate_sql_string(query, sql_to_string_config)

    def execute_frame(
            self,
            result_handler: ResultHandler[R],
            chunk_size: PyLegendOptional[int] = None
    ) -> R:
        from pylegend.core.tds.pandas_api.frames.pandas_api_input_tds_frame import (
            PandasApiInputTdsFrame,
            PandasApiExecutableInputTdsFrame
        )
        tds_frames = self.get_all_tds_frames()
        input_frames = [x for x in tds_frames if isinstance(x, PandasApiInputTdsFrame)]

        non_exec_frames = [x for x in input_frames if not isinstance(x, PandasApiExecutableInputTdsFrame)]
        if non_exec_frames:
            raise ValueError(
                "Cannot execute frame as its built on top of non-executable input frames: [" +
                (", ".join([str(f) for f in non_exec_frames]) + "]")
            )

        exec_frames = [x for x in input_frames if isinstance(x, PandasApiExecutableInputTdsFrame)]

        all_legend_clients = []
        for e in exec_frames:
            c = e.get_legend_client()
            if c not in all_legend_clients:
                all_legend_clients.append(c)
        if len(all_legend_clients) > 1:
            raise ValueError(
                "Found tds frames with multiple legend_clients (which is not supported): [" +
                (", ".join([str(f) for f in all_legend_clients]) + "]")
            )
        legend_client = all_legend_clients[0]
        result = legend_client.execute_sql_string(self.to_sql_query(), chunk_size=chunk_size)
        return result_handler.handle_result(self, result)

    def execute_frame_to_string(
            self,
            chunk_size: PyLegendOptional[int] = None
    ) -> str:
        return self.execute_frame(ToStringResultHandler(), chunk_size)

    def execute_frame_to_pandas_df(
            self,
            chunk_size: PyLegendOptional[int] = None,
            pandas_df_read_config: PandasDfReadConfig = PandasDfReadConfig()
    ) -> pd.DataFrame:
        return self.execute_frame(
            ToPandasDfResultHandler(pandas_df_read_config),
            chunk_size
        )



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta
from enum import Enum
from pylegend.core.language import (
    PyLegendPrimitive,
    PyLegendBoolean,
    PyLegendString,
    PyLegendNumber,
    PyLegendInteger,
    PyLegendFloat,
    PyLegendDate,
    PyLegendDateTime,
    PyLegendStrictDate,
    PyLegendColumnExpression,
    PyLegendExpressionIntegerReturn,
    PyLegendExpressionFloatReturn,
    convert_literal_to_literal_expression,
)
from pylegend._typing import (
    PyLegendSequence,
    PyLegendOptional,
    PyLegendList,
    PyLegendDict,
    PyLegendUnion,
)
from pylegend.core.language.shared.helpers import escape_column_name
from pylegend.core.sql.metamodel import (
    QuerySpecification,
    Expression,
    SingleColumn,
    SortItem,
    SortItemOrdering,
    SortItemNullOrdering,
    Window,
    FunctionCall,
    QualifiedName,
    IntegerLiteral,
    WindowFrame,
    WindowFrameMode,
    FrameBound,
    FrameBoundType,
    StringLiteral
)
from pylegend.core.tds.tds_frame import FrameToSqlConfig, FrameToPureConfig
from typing import TYPE_CHECKING

__all__: PyLegendSequence[str] = [
    "LegendQLApiPrimitive",
    "LegendQLApiBoolean",
    "LegendQLApiString",
    "LegendQLApiNumber",
    "LegendQLApiInteger",
    "LegendQLApiFloat",
    "LegendQLApiDate",
    "LegendQLApiDateTime",
    "LegendQLApiStrictDate",
    "LegendQLApiSortInfo",
    "LegendQLApiSortDirection",
    "LegendQLApiWindow",
    "LegendQLApiPartialFrame",
    "LegendQLApiWindowReference",
    "LegendQLApiWindowFrameBound",
    "LegendQLApiWindowFrameMode",
    "LegendQLApiWindowFrame",
    "LegendQLApiDurationUnit",
    "LegendQLApiWindowFrameBoundType"
]


class LegendQLApiPrimitive(PyLegendPrimitive, metaclass=ABCMeta):
    def ascending(self) -> "LegendQLApiSortInfo":
        val = self.value()
        if isinstance(val, PyLegendColumnExpression):
            return LegendQLApiSortInfo(column_expr=val, direction=LegendQLApiSortDirection.ASC)
        else:
            raise RuntimeError("'ascending' function can only be called on column expressions. "
                               "E.g. - r.col1.ascending() / r['col1'].ascending()\n."
                               f"Found expression type - {type(val)}")

    def descending(self) -> "LegendQLApiSortInfo":
        val = self.value()
        if isinstance(val, PyLegendColumnExpression):
            return LegendQLApiSortInfo(column_expr=val, direction=LegendQLApiSortDirection.DESC)
        else:
            raise RuntimeError("'descending' function can only be called on column expressions. "
                               "E.g. - r.col1.descending() / r['col1'].descending()\n."
                               f"Found expression type - {type(val)}")


class LegendQLApiBoolean(LegendQLApiPrimitive, PyLegendBoolean):
    def __init__(self, expr: PyLegendBoolean):
        PyLegendBoolean.__init__(self, expr.value())


class LegendQLApiString(LegendQLApiPrimitive, PyLegendString):
    def __init__(self, expr: PyLegendString):
        PyLegendString.__init__(self, expr.value())


class LegendQLApiNumber(LegendQLApiPrimitive, PyLegendNumber):
    def __init__(self, expr: PyLegendNumber):
        PyLegendNumber.__init__(self, expr.value())


class LegendQLApiInteger(LegendQLApiPrimitive, PyLegendInteger):
    def __init__(self, expr: PyLegendInteger):
        PyLegendInteger.__init__(self, expr.value())


class LegendQLApiFloat(LegendQLApiPrimitive, PyLegendFloat):
    def __init__(self, expr: PyLegendFloat):
        PyLegendFloat.__init__(self, expr.value())


class LegendQLApiDate(LegendQLApiPrimitive, PyLegendDate):
    def __init__(self, expr: PyLegendDate):
        PyLegendDate.__init__(self, expr.value())


class LegendQLApiDateTime(LegendQLApiPrimitive, PyLegendDateTime):
    def __init__(self, expr: PyLegendDateTime):
        PyLegendDateTime.__init__(self, expr.value())


class LegendQLApiStrictDate(LegendQLApiPrimitive, PyLegendStrictDate):
    def __init__(self, expr: PyLegendStrictDate):
        PyLegendStrictDate.__init__(self, expr.value())


class LegendQLApiSortDirection(Enum):
    ASC = 1,
    DESC = 2


class LegendQLApiSortInfo:
    __column: str
    __direction: LegendQLApiSortDirection

    def __init__(self, column_expr: PyLegendColumnExpression, direction: LegendQLApiSortDirection) -> None:
        self.__column = column_expr.get_column()
        self.__direction = direction

    def to_sql_node(
            self,
            query: QuerySpecification,
            config: FrameToSqlConfig
    ) -> SortItem:
        return SortItem(
            sortKey=self.__find_column_expression(query, config),
            ordering=(SortItemOrdering.ASCENDING if self.__direction == LegendQLApiSortDirection.ASC
                      else SortItemOrdering.DESCENDING),
            nullOrdering=SortItemNullOrdering.UNDEFINED
        )

    def __find_column_expression(self, query: QuerySpecification, config: FrameToSqlConfig) -> Expression:
        db_extension = config.sql_to_string_generator().get_db_extension()
        filtered = [
            s for s in query.select.selectItems
            if (isinstance(s, SingleColumn) and
                s.alias == db_extension.quote_identifier(self.__column))
        ]
        if len(filtered) == 0:
            raise RuntimeError("Cannot find column: " + self.__column)  # pragma: no cover
        return filtered[0].expression

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        func = 'ascending' if self.__direction == LegendQLApiSortDirection.ASC else 'descending'
        return f"{func}(~{escape_column_name(self.__column)})"


class LegendQLApiWindowFrameMode(Enum):
    ROWS = 1
    RANGE = 2


class LegendQLApiWindowFrameBoundType(Enum):
    UNBOUNDED_PRECEDING = 1
    PRECEDING = 2
    CURRENT_ROW = 3
    FOLLOWING = 4
    UNBOUNDED_FOLLOWING = 5

    def to_sql_node(
            self,
            query: QuerySpecification,
            config: FrameToSqlConfig,
    ) -> FrameBoundType:
        mapping = {
            LegendQLApiWindowFrameBoundType.UNBOUNDED_PRECEDING: FrameBoundType.UNBOUNDED_PRECEDING,
            LegendQLApiWindowFrameBoundType.PRECEDING: FrameBoundType.PRECEDING,
            LegendQLApiWindowFrameBoundType.CURRENT_ROW: FrameBoundType.CURRENT_ROW,
            LegendQLApiWindowFrameBoundType.FOLLOWING: FrameBoundType.FOLLOWING,
            LegendQLApiWindowFrameBoundType.UNBOUNDED_FOLLOWING: FrameBoundType.UNBOUNDED_FOLLOWING
        }

        return mapping[self]


class LegendQLApiDurationUnit(Enum):
    YEARS = 1
    MONTHS = 2
    WEEKS = 3
    DAYS = 4
    HOURS = 5
    MINUTES = 6
    SECONDS = 7
    MILLISECONDS = 8
    MICROSECONDS = 9
    NANOSECONDS = 10

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        return self.name

    def to_sql_node(
            self,
            query: QuerySpecification,
            config: FrameToSqlConfig
    ) -> StringLiteral:
        mapping = {
            LegendQLApiDurationUnit.YEARS: "YEAR",
            LegendQLApiDurationUnit.MONTHS: "MONTH",
            LegendQLApiDurationUnit.WEEKS: "WEEK",
            LegendQLApiDurationUnit.DAYS: "DAY",
            LegendQLApiDurationUnit.HOURS: "HOUR",
            LegendQLApiDurationUnit.MINUTES: "MINUTE",
            LegendQLApiDurationUnit.SECONDS: "SECOND",
            LegendQLApiDurationUnit.MILLISECONDS: "MILLISECOND",
            LegendQLApiDurationUnit.MICROSECONDS: "MICROSECOND",
            LegendQLApiDurationUnit.NANOSECONDS: "NANOSECOND",
        }

        return StringLiteral(mapping[self], quoted=False)

    @classmethod
    def from_string(cls, value: str) -> "LegendQLApiDurationUnit":
        try:
            return cls[value.upper()]
        except KeyError:
            raise ValueError(
                f"Invalid duration unit '{value}'. "
                f"Supported values: {[u.name.lower() for u in cls]}"
            )


class LegendQLApiWindowFrameBound:
    __bound_type: LegendQLApiWindowFrameBoundType
    __row_offset: PyLegendOptional[PyLegendUnion[int, float]]
    __duration_unit: PyLegendOptional[LegendQLApiDurationUnit]

    def __init__(
            self,
            bound_type: LegendQLApiWindowFrameBoundType,
            row_offset: PyLegendOptional[PyLegendUnion[int, float]] = None,
            duration_unit: PyLegendOptional[LegendQLApiDurationUnit] = None,
    ) -> None:
        if bound_type in (
                LegendQLApiWindowFrameBoundType.PRECEDING,
                LegendQLApiWindowFrameBoundType.FOLLOWING,
        ) and row_offset is None:
            raise ValueError(f"row_offset must be provided for bound_type {bound_type.name}")

        if bound_type not in (
                LegendQLApiWindowFrameBoundType.PRECEDING,
                LegendQLApiWindowFrameBoundType.FOLLOWING,
        ) and row_offset is not None:
            raise ValueError(f"row_offset is not allowed for bound_type {bound_type.name}")

        self.__bound_type = bound_type
        self.__row_offset = row_offset
        self.__duration_unit = duration_unit

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        if (self.__bound_type == LegendQLApiWindowFrameBoundType.UNBOUNDED_FOLLOWING
                or self.__bound_type == LegendQLApiWindowFrameBoundType.UNBOUNDED_PRECEDING):
            return "unbounded()"

        elif self.__bound_type == LegendQLApiWindowFrameBoundType.CURRENT_ROW:
            expr = "0"

        else:
            expr = convert_literal_to_literal_expression(self.__row_offset).to_pure_expression(
                config) if self.__row_offset is not None else ""

        if self.__duration_unit is not None:
            expr += f", DurationUnit.{self.__duration_unit.to_pure_expression(config)}"

        return expr

    def to_sql_node(
            self,
            query: QuerySpecification,
            config: FrameToSqlConfig,
    ) -> FrameBound:
        value = (convert_literal_to_literal_expression(abs(self.__row_offset))
                 .to_sql_expression({"w": query}, config)) \
            if self.__row_offset is not None else None

        frame_bound_type = self.__bound_type.to_sql_node(query, config)

        duration_unit = self.__duration_unit.to_sql_node(
            query,
            config) if self.__duration_unit is not None else None

        return FrameBound(frame_bound_type, value, duration_unit)


class LegendQLApiWindowFrame:
    __mode: LegendQLApiWindowFrameMode
    __start_bound: LegendQLApiWindowFrameBound
    __end_bound: LegendQLApiWindowFrameBound

    def __init__(
            self,
            mode: LegendQLApiWindowFrameMode,
            start_bound: LegendQLApiWindowFrameBound,
            end_bound: LegendQLApiWindowFrameBound,
    ) -> None:
        self.__mode = mode
        self.__start_bound = start_bound
        self.__end_bound = end_bound

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        mode_str = "rows" if self.__mode == LegendQLApiWindowFrameMode.ROWS else "_range"
        start_expr = self.__start_bound.to_pure_expression(config)
        end_expr = self.__end_bound.to_pure_expression(config)

        return f"{mode_str}({start_expr}, {end_expr})"

    def to_sql_node(
            self,
            query: QuerySpecification,
            config: FrameToSqlConfig
    ) -> WindowFrame:
        return WindowFrame(
            mode=WindowFrameMode.ROWS if self.__mode == LegendQLApiWindowFrameMode.ROWS else WindowFrameMode.RANGE,
            start=self.__start_bound.to_sql_node(query, config),
            end=self.__end_bound.to_sql_node(query, config),
        )


class LegendQLApiWindow:
    __partition_by: PyLegendOptional[PyLegendList[str]]
    __order_by: PyLegendOptional[PyLegendList[LegendQLApiSortInfo]]
    __frame: PyLegendOptional[LegendQLApiWindowFrame]

    def __init__(
            self,
            partition_by: PyLegendOptional[PyLegendList[str]] = None,
            order_by: PyLegendOptional[PyLegendList[LegendQLApiSortInfo]] = None,
            frame: PyLegendOptional[LegendQLApiWindowFrame] = None
    ) -> None:
        self.__partition_by = partition_by
        self.__order_by = order_by
        self.__frame = frame

    def get_partition_by(self) -> PyLegendOptional[PyLegendList[str]]:
        return self.__partition_by

    def get_order_by(self) -> PyLegendOptional[PyLegendList[LegendQLApiSortInfo]]:
        return self.__order_by

    def get_frame(self) -> PyLegendOptional[LegendQLApiWindowFrame]:
        return self.__frame

    def to_sql_node(
            self,
            query: QuerySpecification,
            config: FrameToSqlConfig
    ) -> Window:
        return Window(
            windowRef=None,
            partitions=(
                [] if self.__partition_by is None else
                [LegendQLApiWindow.__find_column_expression(query, col, config) for col in self.__partition_by]
            ),
            orderBy=(
                [] if self.__order_by is None else
                [sort_info.to_sql_node(query, config) for sort_info in self.__order_by]
            ),
            windowFrame=(
                None if self.__frame is None else
                self.__frame.to_sql_node(query, config)
            ),
        )

    @staticmethod
    def __find_column_expression(query: QuerySpecification, col: str, config: FrameToSqlConfig) -> Expression:
        db_extension = config.sql_to_string_generator().get_db_extension()
        filtered = [
            s for s in query.select.selectItems
            if (isinstance(s, SingleColumn) and
                s.alias == db_extension.quote_identifier(col))
        ]
        if len(filtered) == 0:
            raise RuntimeError("Cannot find column: " + col)  # pragma: no cover
        return filtered[0].expression

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        partitions_str = (
            "[]" if self.__partition_by is None or len(self.__partition_by) == 0
            else "~[" + (', '.join(map(escape_column_name, self.__partition_by))) + "]"
        )
        sorts_str = (
            "[]" if self.__order_by is None or len(self.__order_by) == 0
            else "[" + (', '.join([s.to_pure_expression(config) for s in self.__order_by])) + "]"
        )

        frame_str = f", {self.__frame.to_pure_expression(config)}" if self.__frame else ""

        return f"over({partitions_str}, {sorts_str}{frame_str})"


class LegendQLApiPartialFrame:
    if TYPE_CHECKING:
        from pylegend.core.tds.legendql_api.frames.legendql_api_base_tds_frame import LegendQLApiBaseTdsFrame
        from pylegend.core.language.legendql_api.legendql_api_tds_row import LegendQLApiTdsRow

    __base_frame: "LegendQLApiBaseTdsFrame"
    __var_name: str

    def __init__(self, base_frame: "LegendQLApiBaseTdsFrame", var_name: str) -> None:
        self.__base_frame = base_frame
        self.__var_name = var_name

    def row_number(
            self,
            row: "LegendQLApiTdsRow"
    ) -> PyLegendInteger:
        return PyLegendInteger(LegendQLApiRowNumberExpression(self, row))

    def rank(
            self,
            window: "LegendQLApiWindowReference",
            row: "LegendQLApiTdsRow"
    ) -> PyLegendInteger:
        return PyLegendInteger(LegendQLApiRankExpression(self, window, row))

    def dense_rank(
            self,
            window: "LegendQLApiWindowReference",
            row: "LegendQLApiTdsRow"
    ) -> PyLegendInteger:
        return PyLegendInteger(LegendQLApiDenseRankExpression(self, window, row))

    def percent_rank(
            self,
            window: "LegendQLApiWindowReference",
            row: "LegendQLApiTdsRow"
    ) -> PyLegendFloat:
        return PyLegendFloat(LegendQLApiPercentRankExpression(self, window, row))

    def cume_dist(
            self,
            window: "LegendQLApiWindowReference",
            row: "LegendQLApiTdsRow"
    ) -> PyLegendFloat:
        return PyLegendFloat(LegendQLApiCumeDistExpression(self, window, row))

    def ntile(
            self,
            row: "LegendQLApiTdsRow",
            num_buckets: int
    ) -> PyLegendInteger:
        return PyLegendInteger(LegendQLApiNtileExpression(self, row, num_buckets))

    def lead(
            self,
            row: "LegendQLApiTdsRow"
    ) -> "LegendQLApiTdsRow":
        from pylegend.core.language.legendql_api.legendql_api_tds_row import LegendQLApiLeadRow
        return LegendQLApiLeadRow(self, row)

    def lag(
            self,
            row: "LegendQLApiTdsRow"
    ) -> "LegendQLApiTdsRow":
        from pylegend.core.language.legendql_api.legendql_api_tds_row import LegendQLApiLagRow
        return LegendQLApiLagRow(self, row)

    def first(
            self,
            window: "LegendQLApiWindowReference",
            row: "LegendQLApiTdsRow"
    ) -> "LegendQLApiTdsRow":
        from pylegend.core.language.legendql_api.legendql_api_tds_row import LegendQLApiFirstRow
        return LegendQLApiFirstRow(self, window, row)

    def last(
            self,
            window: "LegendQLApiWindowReference",
            row: "LegendQLApiTdsRow"
    ) -> "LegendQLApiTdsRow":
        from pylegend.core.language.legendql_api.legendql_api_tds_row import LegendQLApiLastRow
        return LegendQLApiLastRow(self, window, row)

    def nth(
            self,
            window: "LegendQLApiWindowReference",
            row: "LegendQLApiTdsRow",
            offset: int
    ) -> "LegendQLApiTdsRow":
        from pylegend.core.language.legendql_api.legendql_api_tds_row import LegendQLApiNthRow
        return LegendQLApiNthRow(self, window, row, offset)

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        return f"${self.__var_name}"

    def get_base_frame(self) -> "LegendQLApiBaseTdsFrame":
        return self.__base_frame


class LegendQLApiWindowReference:
    __window: LegendQLApiWindow
    __var_name: str

    def __init__(self, window: LegendQLApiWindow, var_name: str) -> None:
        self.__window = window
        self.__var_name = var_name

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        return f"${self.__var_name}"


class LegendQLApiRowNumberExpression(PyLegendExpressionIntegerReturn):
    if TYPE_CHECKING:
        from pylegend.core.language.legendql_api.legendql_api_tds_row import LegendQLApiTdsRow

    __partial_frame: LegendQLApiPartialFrame
    __row: "LegendQLApiTdsRow"

    def __init__(
            self,
            partial_frame: LegendQLApiPartialFrame,
            row: "LegendQLApiTdsRow"
    ) -> None:
        self.__partial_frame = partial_frame
        self.__row = row

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return FunctionCall(
            name=QualifiedName(parts=["row_number"]), distinct=False, arguments=[], filter_=None, window=None
        )

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        return f"{self.__partial_frame.to_pure_expression(config)}->rowNumber({self.__row.to_pure_expression(config)})"


class LegendQLApiRankExpression(PyLegendExpressionIntegerReturn):
    if TYPE_CHECKING:
        from pylegend.core.language.legendql_api.legendql_api_tds_row import LegendQLApiTdsRow

    __partial_frame: LegendQLApiPartialFrame
    __window_ref: "LegendQLApiWindowReference"
    __row: "LegendQLApiTdsRow"

    def __init__(
            self,
            partial_frame: LegendQLApiPartialFrame,
            window_ref: "LegendQLApiWindowReference",
            row: "LegendQLApiTdsRow"
    ) -> None:
        self.__partial_frame = partial_frame
        self.__window_ref = window_ref
        self.__row = row

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return FunctionCall(
            name=QualifiedName(parts=["rank"]), distinct=False, arguments=[], filter_=None, window=None
        )

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        return (f"{self.__partial_frame.to_pure_expression(config)}->rank("
                f"{self.__window_ref.to_pure_expression(config)}, {self.__row.to_pure_expression(config)})")


class LegendQLApiDenseRankExpression(PyLegendExpressionIntegerReturn):
    if TYPE_CHECKING:
        from pylegend.core.language.legendql_api.legendql_api_tds_row import LegendQLApiTdsRow

    __partial_frame: LegendQLApiPartialFrame
    __window_ref: "LegendQLApiWindowReference"
    __row: "LegendQLApiTdsRow"

    def __init__(
            self,
            partial_frame: LegendQLApiPartialFrame,
            window_ref: "LegendQLApiWindowReference",
            row: "LegendQLApiTdsRow"
    ) -> None:
        self.__partial_frame = partial_frame
        self.__window_ref = window_ref
        self.__row = row

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return FunctionCall(
            name=QualifiedName(parts=["dense_rank"]), distinct=False, arguments=[], filter_=None, window=None
        )

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        return (f"{self.__partial_frame.to_pure_expression(config)}->denseRank("
                f"{self.__window_ref.to_pure_expression(config)}, {self.__row.to_pure_expression(config)})")


class LegendQLApiPercentRankExpression(PyLegendExpressionFloatReturn):
    if TYPE_CHECKING:
        from pylegend.core.language.legendql_api.legendql_api_tds_row import LegendQLApiTdsRow

    __partial_frame: LegendQLApiPartialFrame
    __window_ref: "LegendQLApiWindowReference"
    __row: "LegendQLApiTdsRow"

    def __init__(
            self,
            partial_frame: LegendQLApiPartialFrame,
            window_ref: "LegendQLApiWindowReference",
            row: "LegendQLApiTdsRow"
    ) -> None:
        self.__partial_frame = partial_frame
        self.__window_ref = window_ref
        self.__row = row

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return FunctionCall(
            name=QualifiedName(parts=["percent_rank"]), distinct=False, arguments=[], filter_=None, window=None
        )

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        return (f"{self.__partial_frame.to_pure_expression(config)}->percentRank("
                f"{self.__window_ref.to_pure_expression(config)}, {self.__row.to_pure_expression(config)})")


class LegendQLApiCumeDistExpression(PyLegendExpressionFloatReturn):
    if TYPE_CHECKING:
        from pylegend.core.language.legendql_api.legendql_api_tds_row import LegendQLApiTdsRow

    __partial_frame: LegendQLApiPartialFrame
    __window_ref: "LegendQLApiWindowReference"
    __row: "LegendQLApiTdsRow"

    def __init__(
            self,
            partial_frame: LegendQLApiPartialFrame,
            window_ref: "LegendQLApiWindowReference",
            row: "LegendQLApiTdsRow"
    ) -> None:
        self.__partial_frame = partial_frame
        self.__window_ref = window_ref
        self.__row = row

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return FunctionCall(
            name=QualifiedName(parts=["cume_dist"]), distinct=False, arguments=[], filter_=None, window=None
        )

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        return (f"{self.__partial_frame.to_pure_expression(config)}->cumulativeDistribution("
                f"{self.__window_ref.to_pure_expression(config)}, {self.__row.to_pure_expression(config)})")


class LegendQLApiNtileExpression(PyLegendExpressionIntegerReturn):
    if TYPE_CHECKING:
        from pylegend.core.language.legendql_api.legendql_api_tds_row import LegendQLApiTdsRow

    __partial_frame: LegendQLApiPartialFrame
    __row: "LegendQLApiTdsRow"
    __num_buckets: int

    def __init__(
            self,
            partial_frame: LegendQLApiPartialFrame,
            row: "LegendQLApiTdsRow",
            num_buckets: int
    ) -> None:
        self.__partial_frame = partial_frame
        self.__row = row
        self.__num_buckets = num_buckets

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return FunctionCall(
            name=QualifiedName(parts=["ntile"]),
            distinct=False,
            arguments=[IntegerLiteral(self.__num_buckets)],
            filter_=None,
            window=None
        )

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        return (f"{self.__partial_frame.to_pure_expression(config)}->ntile({self.__row.to_pure_expression(config)}, "
                f"{self.__num_buckets})")



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TYPE_CHECKING

import pandas as pd

from pylegend._typing import (
    PyLegendDict,
)
from pylegend._typing import (
    PyLegendSequence,
    PyLegendOptional,
    PyLegendTypeVar
)
from pylegend.core.language.pandas_api.pandas_api_tds_row import PandasApiTdsRow
from pylegend.core.language.shared.column_expressions import PyLegendColumnExpression
from pylegend.core.language.shared.expression import (
    PyLegendExpressionBooleanReturn,
    PyLegendExpressionStringReturn,
    PyLegendExpressionNumberReturn,
    PyLegendExpressionIntegerReturn,
    PyLegendExpressionFloatReturn,
    PyLegendExpressionDateReturn,
    PyLegendExpressionDateTimeReturn,
    PyLegendExpressionStrictDateReturn,
)
from pylegend.core.language.shared.primitives.boolean import PyLegendBoolean
from pylegend.core.language.shared.primitives.date import PyLegendDate
from pylegend.core.language.shared.primitives.datetime import PyLegendDateTime
from pylegend.core.language.shared.primitives.float import PyLegendFloat
from pylegend.core.language.shared.primitives.integer import PyLegendInteger
from pylegend.core.language.shared.primitives.number import PyLegendNumber
from pylegend.core.language.shared.primitives.primitive import PyLegendPrimitive
from pylegend.core.language.shared.primitives.strictdate import PyLegendStrictDate
from pylegend.core.language.shared.primitives.string import PyLegendString
from pylegend.core.sql.metamodel import (
    Expression,
)
from pylegend.core.sql.metamodel import QuerySpecification
from pylegend.core.tds.abstract.frames.base_tds_frame import BaseTdsFrame
from pylegend.core.tds.result_handler import ResultHandler
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToPureConfig
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.extensions.tds.result_handler import PandasDfReadConfig

if TYPE_CHECKING:
    from pylegend.core.tds.pandas_api.frames.pandas_api_tds_frame import PandasApiTdsFrame

__all__: PyLegendSequence[str] = [
    "Series"
]

R = PyLegendTypeVar('R')


class Series(PyLegendColumnExpression, PyLegendPrimitive, BaseTdsFrame):
    def __init__(self, base_frame: "PandasApiTdsFrame", column: str):
        row = PandasApiTdsRow.from_tds_frame("c", base_frame)
        PyLegendColumnExpression.__init__(self, row=row, column=column)

        self.__base_frame = base_frame
        self._filtered_frame = base_frame.filter(items=[column])

    def value(self) -> PyLegendColumnExpression:
        return self

    def get_base_frame(self) -> "PandasApiTdsFrame":
        return self.__base_frame

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        return super().to_sql_expression(frame_name_to_base_query_map, config)

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        return super().to_pure_expression(config)

    def columns(self) -> PyLegendSequence[TdsColumn]:
        return self._filtered_frame.columns()

    def to_sql_query(self, config: FrameToSqlConfig = FrameToSqlConfig()) -> str:
        return self._filtered_frame.to_sql_query(config)

    def to_pure_query(self, config: FrameToPureConfig = FrameToPureConfig()) -> str:
        return self._filtered_frame.to_pure_query(config)

    def execute_frame(
            self,
            result_handler: ResultHandler[R],
            chunk_size: PyLegendOptional[int] = None
    ) -> R:
        return self._filtered_frame.execute_frame(result_handler, chunk_size)  # pragma: no cover

    def execute_frame_to_string(
            self,
            chunk_size: PyLegendOptional[int] = None
    ) -> str:
        return self._filtered_frame.execute_frame_to_string(chunk_size)

    def execute_frame_to_pandas_df(
            self,
            chunk_size: PyLegendOptional[int] = None,
            pandas_df_read_config: PandasDfReadConfig = PandasDfReadConfig()
    ) -> pd.DataFrame:
        return self._filtered_frame.execute_frame_to_pandas_df(chunk_size, pandas_df_read_config)  # pragma: no cover

    def to_sql_query_object(self, config: FrameToSqlConfig) -> QuerySpecification:
        return self._filtered_frame.to_sql_query_object(config)  # type: ignore

    def to_pure(self, config: FrameToPureConfig) -> str:
        return self._filtered_frame.to_pure(config)  # type: ignore

    def get_all_tds_frames(self) -> PyLegendSequence["BaseTdsFrame"]:
        return self._filtered_frame.get_all_tds_frames()  # type: ignore


class BooleanSeries(Series, PyLegendBoolean, PyLegendExpressionBooleanReturn):  # type: ignore
    def __init__(self, base_frame: "PandasApiTdsFrame", column: str):
        super().__init__(base_frame, column)  # pragma: no cover (Boolean column not supported in PURE)
        PyLegendBoolean.__init__(self, self)  # pragma: no cover (Boolean column not supported in PURE)


class StringSeries(Series, PyLegendString, PyLegendExpressionStringReturn):  # type: ignore
    def __init__(self, base_frame: "PandasApiTdsFrame", column: str):
        super().__init__(base_frame, column)
        PyLegendString.__init__(self, self)


class NumberSeries(Series, PyLegendNumber, PyLegendExpressionNumberReturn):  # type: ignore
    def __init__(self, base_frame: "PandasApiTdsFrame", column: str):
        super().__init__(base_frame, column)
        PyLegendNumber.__init__(self, self)


class IntegerSeries(NumberSeries, PyLegendInteger, PyLegendExpressionIntegerReturn):  # type: ignore
    def __init__(self, base_frame: "PandasApiTdsFrame", column: str):
        super().__init__(base_frame, column)
        PyLegendInteger.__init__(self, self)


class FloatSeries(NumberSeries, PyLegendFloat, PyLegendExpressionFloatReturn):  # type: ignore
    def __init__(self, base_frame: "PandasApiTdsFrame", column: str):
        super().__init__(base_frame, column)
        PyLegendFloat.__init__(self, self)


class DateSeries(Series, PyLegendDate, PyLegendExpressionDateReturn):  # type: ignore
    def __init__(self, base_frame: "PandasApiTdsFrame", column: str):
        super().__init__(base_frame, column)
        PyLegendDate.__init__(self, self)


class DateTimeSeries(DateSeries, PyLegendDateTime, PyLegendExpressionDateTimeReturn):  # type: ignore
    def __init__(self, base_frame: "PandasApiTdsFrame", column: str):
        super().__init__(base_frame, column)
        PyLegendDateTime.__init__(self, self)


class StrictDateSeries(DateSeries, PyLegendStrictDate, PyLegendExpressionStrictDateReturn):  # type: ignore
    def __init__(self, base_frame: "PandasApiTdsFrame", column: str):
        super().__init__(base_frame, column)
        PyLegendStrictDate.__init__(self, self)



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from datetime import date, datetime
from typing import TYPE_CHECKING

from typing_extensions import Concatenate

try:
    from typing import ParamSpec
except Exception:
    from typing_extensions import ParamSpec  # type: ignore

from pylegend._typing import (
    PyLegendCallable,
    PyLegendSequence,
    PyLegendUnion,
    PyLegendOptional,
    PyLegendList,
    PyLegendSet,
    PyLegendTuple,
    PyLegendDict
)
from pylegend.core.language import (
    PyLegendPrimitive,
)
from pylegend.core.language.pandas_api.pandas_api_aggregate_specification import PyLegendAggInput
from pylegend.core.language.pandas_api.pandas_api_tds_row import PandasApiTdsRow
from pylegend.core.language.shared.primitives.boolean import PyLegendBoolean
from pylegend.core.language.shared.primitives.integer import PyLegendInteger
from pylegend.core.language.shared.primitives.primitive import PyLegendPrimitiveOrPythonPrimitive
from pylegend.core.language.shared.tds_row import AbstractTdsRow
from pylegend.core.tds.tds_frame import PyLegendTdsFrame

if TYPE_CHECKING:
    from pylegend.core.language.pandas_api.pandas_api_series import Series
    from pylegend.core.tds.pandas_api.frames.pandas_api_groupby_tds_frame import PandasApiGroupbyTdsFrame

__all__: PyLegendSequence[str] = [
    "PandasApiTdsFrame"
]

P = ParamSpec("P")


class PandasApiTdsFrame(PyLegendTdsFrame):

    @abstractmethod
    def __getitem__(
            self,
            key: PyLegendUnion[str, PyLegendList[str], PyLegendBoolean]
    ) -> PyLegendUnion["PandasApiTdsFrame", "Series"]:
        pass  # pragma: no cover

    @abstractmethod
    def __setitem__(
            self,
            key: str,
            value: PyLegendUnion["Series", PyLegendPrimitiveOrPythonPrimitive]
    ) -> None:
        pass  # pragma: no cover

    @abstractmethod
    def assign(
            self,
            **kwargs: PyLegendCallable[
                [PandasApiTdsRow],
                PyLegendUnion[int, float, bool, str, date, datetime, PyLegendPrimitive]
            ],
    ) -> "PandasApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def filter(
            self,
            items: PyLegendOptional[PyLegendList[str]] = None,
            like: PyLegendOptional[str] = None,
            regex: PyLegendOptional[str] = None,
            axis: PyLegendOptional[PyLegendUnion[str, int, PyLegendInteger]] = None
    ) -> "PandasApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def sort_values(
            self,
            by: PyLegendUnion[str, PyLegendList[str]],
            axis: PyLegendUnion[str, int] = 0,
            ascending: PyLegendUnion[bool, PyLegendList[bool]] = True,
            inplace: bool = False,
            kind: PyLegendOptional[str] = None,
            na_position: str = 'last',
            ignore_index: bool = True,
            key: PyLegendOptional[PyLegendCallable[[AbstractTdsRow], AbstractTdsRow]] = None
    ) -> "PandasApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def truncate(
            self,
            before: PyLegendUnion[date, str, int, None] = 0,
            after: PyLegendUnion[date, str, int, None] = None,
            axis: PyLegendUnion[str, int] = 0,
            copy: bool = True
    ) -> "PandasApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def drop(
            self,
            labels: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str], PyLegendSet[str]]] = None,
            axis: PyLegendUnion[str, int, PyLegendInteger] = 1,
            index: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str], PyLegendSet[str]]] = None,
            columns: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str], PyLegendSet[str]]] = None,
            level: PyLegendOptional[PyLegendUnion[int, PyLegendInteger, str]] = None,
            inplace: PyLegendUnion[bool, PyLegendBoolean] = False,
            errors: str = "raise",
    ) -> "PandasApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def aggregate(
            self,
            func: PyLegendAggInput,
            axis: PyLegendUnion[int, str] = 0,
            *args: PyLegendPrimitiveOrPythonPrimitive,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def agg(
            self,
            func: PyLegendAggInput,
            axis: PyLegendUnion[int, str] = 0,
            *args: PyLegendPrimitiveOrPythonPrimitive,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def merge(
            self,
            other: "PandasApiTdsFrame",
            how: PyLegendOptional[str] = "inner",
            on: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str]]] = None,
            left_on: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str]]] = None,
            right_on: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str]]] = None,
            left_index: PyLegendOptional[bool] = False,
            right_index: PyLegendOptional[bool] = False,
            sort: PyLegendOptional[bool] = False,
            suffixes: PyLegendOptional[
                PyLegendUnion[
                    PyLegendTuple[PyLegendUnion[str, None], PyLegendUnion[str, None]],
                    PyLegendList[PyLegendUnion[str, None]],
                ]
            ] = ("_x", "_y"),
            indicator: PyLegendOptional[PyLegendUnion[bool, str]] = False,
            validate: PyLegendOptional[str] = None
    ) -> "PandasApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def join(
            self,
            other: "PandasApiTdsFrame",
            on: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str]]] = None,
            how: PyLegendOptional[str] = "left",
            lsuffix: str = "",
            rsuffix: str = "",
            sort: PyLegendOptional[bool] = False,
            validate: PyLegendOptional[str] = None
    ) -> "PandasApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def rename(
            self,
            mapper: PyLegendOptional[PyLegendUnion[PyLegendDict[str, str], PyLegendCallable[[str], str]]] = None,
            index: PyLegendOptional[PyLegendUnion[PyLegendDict[str, str], PyLegendCallable[[str], str]]] = None,
            columns: PyLegendOptional[PyLegendUnion[PyLegendDict[str, str], PyLegendCallable[[str], str]]] = None,
            axis: PyLegendUnion[str, int] = 1,
            inplace: PyLegendUnion[bool] = False,
            copy: PyLegendUnion[bool] = True,
            level: PyLegendOptional[PyLegendUnion[int, str]] = None,
            errors: str = "ignore",
    ) -> "PandasApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def groupby(
            self,
            by: PyLegendUnion[str, PyLegendList[str]],
            level: PyLegendOptional[PyLegendUnion[str, int, PyLegendList[str]]] = None,
            as_index: bool = False,
            sort: bool = True,
            group_keys: bool = False,
            observed: bool = False,
            dropna: bool = False,
    ) -> "PandasApiGroupbyTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def sum(
            self,
            axis: PyLegendUnion[int, str] = 0,
            skipna: bool = True,
            numeric_only: bool = False,
            min_count: int = 0,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def mean(
            self,
            axis: PyLegendUnion[int, str] = 0,
            skipna: bool = True,
            numeric_only: bool = False,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def min(
            self,
            axis: PyLegendUnion[int, str] = 0,
            skipna: bool = True,
            numeric_only: bool = False,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def max(
            self,
            axis: PyLegendUnion[int, str] = 0,
            skipna: bool = True,
            numeric_only: bool = False,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def std(
            self,
            axis: PyLegendUnion[int, str] = 0,
            skipna: bool = True,
            ddof: int = 1,
            numeric_only: bool = False,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def var(
            self,
            axis: PyLegendUnion[int, str] = 0,
            skipna: bool = True,
            ddof: int = 1,
            numeric_only: bool = False,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def count(
            self,
            axis: PyLegendUnion[int, str] = 0,
            numeric_only: bool = False,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def apply(
            self,
            func: PyLegendUnion[
                PyLegendCallable[Concatenate["Series", P], PyLegendPrimitiveOrPythonPrimitive],
                str
            ],
            axis: PyLegendUnion[int, str] = 0,
            raw: bool = False,
            result_type: PyLegendOptional[str] = None,
            args: PyLegendTuple[PyLegendPrimitiveOrPythonPrimitive, ...] = (),
            by_row: PyLegendUnion[bool, str] = "compat",
            engine: str = "python",
            engine_kwargs: PyLegendOptional[PyLegendDict[str, PyLegendPrimitiveOrPythonPrimitive]] = None,
            **kwargs: PyLegendPrimitiveOrPythonPrimitive
    ) -> "PandasApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def head(self, n: int = 5) -> "PandasApiTdsFrame":
        pass  # pragma: no cover

    @property
    @abstractmethod
    def shape(self) -> PyLegendTuple[int, int]:
        pass  # pragma: no cover

    @abstractmethod
    def dropna(
            self,
            axis: PyLegendUnion[int, str] = 0,
            how: str = "any",
            thresh: PyLegendOptional[int] = None,
            subset: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str]]] = None,
            inplace: bool = False,
            ignore_index: bool = False
    ) -> "PandasApiTdsFrame":
        pass  # pragma: no cover

    @abstractmethod
    def fillna(
            self,
            value: PyLegendUnion[
                int, float, str, bool, date, datetime,
                PyLegendDict[str, PyLegendUnion[int, float, str, bool, date, datetime]]
            ] = None,  # type: ignore
            axis: PyLegendOptional[PyLegendUnion[int, str]] = 0,
            inplace: bool = False,
            limit: PyLegendOptional[int] = None
    ) -> "PandasApiTdsFrame":
        pass  # pragma: no cover



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendList,
    PyLegendSequence
)
from pylegend.core.tds.legendql_api.frames.legendql_api_applied_function_tds_frame import LegendQLApiAppliedFunction
from pylegend.core.sql.metamodel import (
    QuerySpecification,
    Union,
    AliasedRelation,
    SingleColumn,
    Select,
    QualifiedNameReference,
    QualifiedName,
    TableSubquery,
    Query,
)
from pylegend.core.tds.sql_query_helpers import create_sub_query
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.tds.tds_frame import FrameToPureConfig
from pylegend.core.tds.legendql_api.frames.legendql_api_tds_frame import LegendQLApiTdsFrame
from pylegend.core.tds.legendql_api.frames.legendql_api_base_tds_frame import LegendQLApiBaseTdsFrame


__all__: PyLegendSequence[str] = [
    "LegendQLApiConcatenateFunction"
]


class LegendQLApiConcatenateFunction(LegendQLApiAppliedFunction):
    __base_frame: LegendQLApiBaseTdsFrame
    __other_frame: LegendQLApiBaseTdsFrame

    @classmethod
    def name(cls) -> str:
        return "concatenate"

    def __init__(self, base_frame: LegendQLApiBaseTdsFrame, other: LegendQLApiTdsFrame) -> None:
        self.__base_frame = base_frame
        if not isinstance(other, LegendQLApiBaseTdsFrame):
            raise ValueError("Expected LegendQLApiBaseTdsFrame")  # pragma: no cover
        self.__other_frame = other

    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        base_query = self.__base_frame.to_sql_query_object(config)
        other_query = self.__other_frame.to_sql_query_object(config)

        db_extension = config.sql_to_string_generator().get_db_extension()
        root_alias = db_extension.quote_identifier("root")
        columns = [db_extension.quote_identifier(c.get_name()) for c in self.__base_frame.columns()]

        return QuerySpecification(
            select=Select(
                selectItems=[
                    SingleColumn(
                        alias=x,
                        expression=QualifiedNameReference(name=QualifiedName(parts=[root_alias, x]))
                    )
                    for x in columns
                ],
                distinct=False
            ),
            from_=[
                AliasedRelation(
                    relation=TableSubquery(
                        query=Query(
                            queryBody=Union(
                                left=create_sub_query(base_query, config, "left"),
                                right=create_sub_query(other_query, config, "right"),
                                distinct=False
                            ),
                            limit=None, offset=None, orderBy=[]
                        )
                    ),
                    alias=root_alias,
                    columnNames=columns
                )
            ],
            where=None,
            groupBy=[],
            having=None,
            orderBy=[],
            limit=None,
            offset=None
        )

    def to_pure(self, config: FrameToPureConfig) -> str:
        return (f"{self.__base_frame.to_pure(config)}{config.separator(1)}"
                f"->concatenate({config.separator(2)}"
                f"{self.__other_frame.to_pure(config.push_indent(2))}"
                f"{config.separator(1)})")

    def base_frame(self) -> LegendQLApiBaseTdsFrame:
        return self.__base_frame

    def tds_frame_parameters(self) -> PyLegendList["LegendQLApiBaseTdsFrame"]:
        return [self.__other_frame]

    def calculate_columns(self) -> PyLegendSequence["TdsColumn"]:
        return [c.copy() for c in self.__base_frame.columns()]

    def validate(self) -> bool:
        base_frame_cols = self.__base_frame.columns()
        other_frame_cols = self.__other_frame.columns()

        if len(base_frame_cols) != len(other_frame_cols):
            cols1 = "[" + ", ".join([str(c) for c in base_frame_cols]) + "]"
            cols2 = "[" + ", ".join([str(c) for c in other_frame_cols]) + "]"
            raise ValueError(
                "Cannot concatenate two Tds Frames with different column counts. \n"
                f"Frame 1 cols - (Count: {len(base_frame_cols)}) - {cols1} \n"
                f"Frame 2 cols - (Count: {len(other_frame_cols)}) - {cols2} \n"
            )

        for i in range(0, len(base_frame_cols)):
            base_col = base_frame_cols[i]
            other_col = other_frame_cols[i]

            if (base_col.get_name() != other_col.get_name()) or (base_col.get_type() != other_col.get_type()):
                raise ValueError(
                    f"Column name/type mismatch when concatenating Tds Frames at index {i}. "
                    f"Frame 1 column - {base_col}, Frame 2 column - {other_col}"
                )

        return True



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendSequence,
)
from pylegend.core.language import (
    PyLegendBoolean,
    PyLegendString,
    PyLegendInteger,
    PyLegendFloat,
    PyLegendNumber,
    PyLegendStrictDate,
    PyLegendDateTime,
    PyLegendDate,
)
from pylegend.core.language.pandas_api.pandas_api_custom_expressions import (
    PandasApiBoolean,
    PandasApiString,
    PandasApiInteger,
    PandasApiFloat,
    PandasApiNumber,
    PandasApiStrictDate,
    PandasApiDateTime,
    PandasApiDate,
    PandasApiPrimitive,
)
from pylegend.core.language.shared.tds_row import AbstractTdsRow
from pylegend.core.tds.tds_frame import PyLegendTdsFrame

__all__: PyLegendSequence[str] = [
    "PandasApiTdsRow",
]


class PandasApiTdsRow(AbstractTdsRow):
    def __init__(self, frame_name: str, frame: PyLegendTdsFrame) -> None:
        super().__init__(frame_name, frame)

    @staticmethod
    def from_tds_frame(frame_name: str, frame: PyLegendTdsFrame) -> "PandasApiTdsRow":
        return PandasApiTdsRow(frame_name=frame_name, frame=frame)

    def __getitem__(self, item: str) -> PandasApiPrimitive:
        res = super().__getitem__(item)
        if isinstance(res, PyLegendBoolean):
            return PandasApiBoolean(res)
        if isinstance(res, PyLegendString):
            return PandasApiString(res)
        if isinstance(res, PyLegendInteger):
            return PandasApiInteger(res)
        if isinstance(res, PyLegendFloat):
            return PandasApiFloat(res)
        if isinstance(res, PyLegendNumber):
            return PandasApiNumber(res)
        if isinstance(res, PyLegendStrictDate):
            return PandasApiStrictDate(res)
        if isinstance(res, PyLegendDateTime):
            return PandasApiDateTime(res)
        if isinstance(res, PyLegendDate):
            return PandasApiDate(res)

        raise RuntimeError(f"Unhandled primitive type {type(res)} in Pandas Api")  # pragma: no cover



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABCMeta
from pylegend._typing import (
    PyLegendSequence,
    PyLegendDict,
    PyLegendCallable,
    PyLegendList,
    PyLegendOptional
)
from pylegend.core.language.shared.expression import (
    PyLegendExpression,
)
from pylegend.core.language.shared.helpers import expr_has_matching_start_and_end_parentheses
from pylegend.core.sql.metamodel import (
    Expression,
    QuerySpecification,
)
from pylegend.core.tds.tds_frame import FrameToSqlConfig
from pylegend.core.tds.tds_frame import FrameToPureConfig

__all__: PyLegendSequence[str] = [
    "PyLegendNaryExpression"
]


class PyLegendNaryExpression(PyLegendExpression, metaclass=ABCMeta):
    __operands: PyLegendList[PyLegendExpression]
    __to_sql_func: PyLegendCallable[
        [PyLegendList[Expression], PyLegendDict[str, QuerySpecification], FrameToSqlConfig],
        Expression
    ]
    __to_pure_func: PyLegendCallable[
        [PyLegendList[str], FrameToPureConfig],
        str
    ]
    __non_nullable: bool
    __operands_non_nullable_flags: PyLegendList[bool]

    def __init__(
            self,
            operands: PyLegendList[PyLegendExpression],
            to_sql_func: PyLegendCallable[
                [PyLegendList[Expression], PyLegendDict[str, QuerySpecification], FrameToSqlConfig],
                Expression
            ],
            to_pure_func: PyLegendCallable[
                [PyLegendList[str], FrameToPureConfig],
                str
            ],
            non_nullable: bool = False,
            operands_non_nullable_flags: PyLegendOptional[PyLegendList[bool]] = None
    ) -> None:
        self.__operands = operands
        self.__to_sql_func = to_sql_func
        self.__to_pure_func = to_pure_func
        self.__non_nullable = non_nullable
        self.__operands_non_nullable_flags = (
            operands_non_nullable_flags
            if operands_non_nullable_flags is not None
            else [False] * len(operands)
        )

    def to_sql_expression(
            self,
            frame_name_to_base_query_map: PyLegendDict[str, QuerySpecification],
            config: FrameToSqlConfig
    ) -> Expression:
        sql_operands = [
            operand.to_sql_expression(frame_name_to_base_query_map, config)
            for operand in self.__operands
        ]
        return self.__to_sql_func(sql_operands, frame_name_to_base_query_map, config)

    def to_pure_expression(self, config: FrameToPureConfig) -> str:
        pure_operands: PyLegendList[str] = []

        for operand, must_be_non_nullable in zip(self.__operands, self.__operands_non_nullable_flags):
            expr = operand.to_pure_expression(config)

            if must_be_non_nullable:
                expr = (
                    expr if operand.is_non_nullable() else
                    f"toOne({expr[1:-1] if expr_has_matching_start_and_end_parentheses(expr) else expr})"
                )

            pure_operands.append(expr)

        return self.__to_pure_func(pure_operands, config)

    def is_non_nullable(self) -> bool:
        return self.__non_nullable



# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from enum import Enum
from pylegend._typing import (
    PyLegendList,
    PyLegendSequence,
    PyLegendOptional
)


__all__: PyLegendSequence[str] = [
    'WindowFrameMode',
    'FrameBoundType',
    'TrimMode',
    'JoinType',
    'LogicalBinaryType',
    'ArithmeticType',
    'SortItemOrdering',
    'SortItemNullOrdering',
    'ComparisonOperator',
    'CurrentTimeType',
    'ExtractField',
    'Node',
    'Statement',
    'Query',
    'Relation',
    'QueryBody',
    'TableSubquery',
    'QuerySpecification',
    'SetOperation',
    'AliasedRelation',
    'Union',
    'Select',
    'SelectItem',
    'AllColumns',
    'SingleColumn',
    'Table',
    'TableFunction',
    'Expression',
    'SubqueryExpression',
    'Literal',
    'LongLiteral',
    'BooleanLiteral',
    'DoubleLiteral',
    'IntegerLiteral',
    'StringLiteral',
    'ArrayLiteral',
    'NullLiteral',
    'IntervalLiteral',
    'NamedArgumentExpression',
    'SortItem',
    'ComparisonExpression',
    'LogicalBinaryExpression',
    'NotExpression',
    'ArithmeticExpression',
    'NegativeExpression',
    'IsNullPredicate',
    'IsNotNullPredicate',
    'CurrentTime',
    'FunctionCall',
    'SimpleCaseExpression',
    'SearchedCaseExpression',
    'WhenClause',
    'Extract',
    'Join',
    'JoinCriteria',
    'JoinOn',
    'JoinUsing',
    'Cast',
    'ColumnType',
    'QualifiedName',
    'QualifiedNameReference',
    'InListExpression',
    'InPredicate',
    'Window',
    'WindowFrame',
    'FrameBound'
]


class WindowFrameMode(Enum):
    RANGE = 1,
    ROWS = 2


class FrameBoundType(Enum):
    UNBOUNDED_PRECEDING = 1,
    PRECEDING = 2,
    CURRENT_ROW = 3,
    FOLLOWING = 4,
    UNBOUNDED_FOLLOWING = 5


class TrimMode(Enum):
    LEADING = 1,
    TRAILING = 2,
    BOTH = 3


class JoinType(Enum):
    CROSS = 1,
    INNER = 2,
    LEFT = 3,
    RIGHT = 4,
    FULL = 5


class LogicalBinaryType(Enum):
    AND = 1,
    OR = 2


class ArithmeticType(Enum):
    ADD = 1,
    SUBTRACT = 2,
    MULTIPLY = 3,
    DIVIDE = 4,
    MODULUS = 5


class SortItemOrdering(Enum):
    ASCENDING = 1,
    DESCENDING = 2


class SortItemNullOrdering(Enum):
    FIRST = 1,
    LAST = 2,
    UNDEFINED = 3


class ComparisonOperator(Enum):
    EQUAL = 1,
    NOT_EQUAL = 2,
    LESS_THAN = 3,
    LESS_THAN_OR_EQUAL = 4,
    GREATER_THAN = 5,
    GREATER_THAN_OR_EQUAL = 6,
    REGEX_MATCH = 7,
    LIKE = 8


class CurrentTimeType(Enum):
    DATE = 1,
    TIME = 2,
    TIMESTAMP = 3


class ExtractField(Enum):
    CENTURY = 1,
    YEAR = 2,
    QUARTER = 3,
    MONTH = 4,
    WEEK = 5,
    DAY = 6,
    DAY_OF_MONTH = 7,
    DAY_OF_WEEK = 8,
    DOW = 9,
    DAY_OF_YEAR = 10,
    DOY = 11,
    HOUR = 12,
    MINUTE = 13,
    SECOND = 14,
    TIMEZONE_HOUR = 15,
    TIMEZONE_MINUTE = 16,
    EPOCH = 17


class Node:
    _type: str

    def __init__(
        self,
        _type: str
    ) -> None:
        self._type = _type


class Statement(Node):

    def __init__(
        self,
        _type: str = "statement"
    ) -> None:
        super().__init__(_type=_type)


class Query(Statement):
    queryBody: "QueryBody"
    limit: "PyLegendOptional[Expression]"
    orderBy: "PyLegendList[SortItem]"
    offset: "PyLegendOptional[Expression]"

    def __init__(
        self,
        queryBody: "QueryBody",
        limit: "PyLegendOptional[Expression]",
        orderBy: "PyLegendList[SortItem]",
        offset: "PyLegendOptional[Expression]"
    ) -> None:
        super().__init__(_type="query")
        self.queryBody = queryBody
        self.limit = limit
        self.orderBy = orderBy
        self.offset = offset


class Relation(Node):

    def __init__(
        self,
        _type: str = "relation"
    ) -> None:
        super().__init__(_type=_type)


class QueryBody(Relation):

    def __init__(
        self,
        _type: str = "queryBody"
    ) -> None:
        super().__init__(_type=_type)


class TableSubquery(QueryBody):
    query: "Query"

    def __init__(
        self,
        query: "Query"
    ) -> None:
        super().__init__(_type="tableSubquery")
        self.query = query


class QuerySpecification(QueryBody):
    select: "Select"
    from_: "PyLegendList[Relation]"
    where: "PyLegendOptional[Expression]"
    groupBy: "PyLegendList[Expression]"
    having: "PyLegendOptional[Expression]"
    orderBy: "PyLegendList[SortItem]"
    limit: "PyLegendOptional[Expression]"
    offset: "PyLegendOptional[Expression]"

    def __init__(
        self,
        select: "Select",
        from_: "PyLegendList[Relation]",
        where: "PyLegendOptional[Expression]",
        groupBy: "PyLegendList[Expression]",
        having: "PyLegendOptional[Expression]",
        orderBy: "PyLegendList[SortItem]",
        limit: "PyLegendOptional[Expression]",
        offset: "PyLegendOptional[Expression]"
    ) -> None:
        super().__init__(_type="querySpecification")
        self.select = select
        self.from_ = from_
        self.where = where
        self.groupBy = groupBy
        self.having = having
        self.orderBy = orderBy
        self.limit = limit
        self.offset = offset


class SetOperation(QueryBody):

    def __init__(
        self,
        _type: str = "setOperation"
    ) -> None:
        super().__init__(_type=_type)


class AliasedRelation(Relation):
    relation: "Relation"
    alias: "str"
    columnNames: "PyLegendList[str]"

    def __init__(
        self,
        relation: "Relation",
        alias: "str",
        columnNames: "PyLegendList[str]"
    ) -> None:
        super().__init__(_type="aliasedRelation")
        self.relation = relation
        self.alias = alias
        self.columnNames = columnNames


class Union(SetOperation):
    left: "Relation"
    right: "Relation"
    distinct: "bool"

    def __init__(
        self,
        left: "Relation",
        right: "Relation",
        distinct: "bool"
    ) -> None:
        super().__init__(_type="union")
        self.left = left
        self.right = right
        self.distinct = distinct


class Select(Node):
    distinct: "bool"
    selectItems: "PyLegendList[SelectItem]"

    def __init__(
        self,
        distinct: "bool",
        selectItems: "PyLegendList[SelectItem]"
    ) -> None:
        super().__init__(_type="select")
        self.distinct = distinct
        self.selectItems = selectItems


class SelectItem(Node):

    def __init__(
        self,
        _type: str = "selectItem"
    ) -> None:
        super().__init__(_type=_type)


class AllColumns(SelectItem):
    prefix: "PyLegendOptional[str]"

    def __init__(
        self,
        prefix: "PyLegendOptional[str]"
    ) -> None:
        super().__init__(_type="allColumns")
        self.prefix = prefix


class SingleColumn(SelectItem):
    alias: "PyLegendOptional[str]"
    expression: "Expression"

    def __init__(
        self,
        alias: "PyLegendOptional[str]",
        expression: "Expression"
    ) -> None:
        super().__init__(_type="singleColumn")
        self.alias = alias
        self.expression = expression


class Table(QueryBody):
    name: "QualifiedName"

    def __init__(
        self,
        name: "QualifiedName"
    ) -> None:
        super().__init__(_type="table")
        self.name = name


class TableFunction(QueryBody):
    functionCall: "FunctionCall"

    def __init__(
        self,
        functionCall: "FunctionCall"
    ) -> None:
        super().__init__(_type="tableFunction")
        self.functionCall = functionCall


class Expression(Node):

    def __init__(
        self,
        _type: str = "expression"
    ) -> None:
        super().__init__(_type=_type)


class SubqueryExpression(Expression):
    query: "Query"

    def __init__(
        self,
        query: "Query"
    ) -> None:
        super().__init__(_type="subqueryExpression")
        self.query = query


class Literal(Expression):

    def __init__(
        self,
        _type: str = "literal"
    ) -> None:
        super().__init__(_type=_type)


class LongLiteral(Literal):
    value: "int"

    def __init__(
        self,
        value: "int"
    ) -> None:
        super().__init__(_type="longLiteral")
        self.value = value


class BooleanLiteral(Literal):
    value: "bool"

    def __init__(
        self,
        value: "bool"
    ) -> None:
        super().__init__(_type="booleanLiteral")
        self.value = value


class DoubleLiteral(Literal):
    value: "float"

    def __init__(
        self,
        value: "float"
    ) -> None:
        super().__init__(_type="doubleLiteral")
        self.value = value


class IntegerLiteral(Literal):
    value: "int"

    def __init__(
        self,
        value: "int"
    ) -> None:
        super().__init__(_type="integerLiteral")
        self.value = value


class StringLiteral(Literal):
    value: "str"
    quoted: "bool"

    def __init__(
        self,
        value: "str",
        quoted: "bool"
    ) -> None:
        super().__init__(_type="stringLiteral")
        self.value = value
        self.quoted = quoted


class ArrayLiteral(Literal):
    values: "PyLegendList[Expression]"

    def __init__(
        self,
        values: "PyLegendList[Expression]"
    ) -> None:
        super().__init__(_type="arrayLiteral")
        self.values = values


class NullLiteral(Literal):

    def __init__(
        self
    ) -> None:
        super().__init__(_type="nullLiteral")


class IntervalLiteral(Literal):
    ago: "PyLegendOptional[bool]"
    years: "PyLegendOptional[int]"
    months: "PyLegendOptional[int]"
    weeks: "PyLegendOptional[int]"
    days: "PyLegendOptional[int]"
    hours: "PyLegendOptional[int]"
    minutes: "PyLegendOptional[int]"
    seconds: "PyLegendOptional[int]"

    def __init__(
        self,
        ago: "PyLegendOptional[bool]",
        years: "PyLegendOptional[int]",
        months: "PyLegendOptional[int]",
        weeks: "PyLegendOptional[int]",
        days: "PyLegendOptional[int]",
        hours: "PyLegendOptional[int]",
        minutes: "PyLegendOptional[int]",
        seconds: "PyLegendOptional[int]"
    ) -> None:
        super().__init__(_type="intervalLiteral")
        self.ago = ago
        self.years = years
        self.months = months
        self.weeks = weeks
        self.days = days
        self.hours = hours
        self.minutes = minutes
        self.seconds = seconds


class NamedArgumentExpression(Expression):
    name: "str"
    expression: "Expression"

    def __init__(
        self,
        name: "str",
        expression: "Expression"
    ) -> None:
        super().__init__(_type="namedArgumentExpression")
        self.name = name
        self.expression = expression


class SortItem(Node):
    sortKey: "Expression"
    ordering: "SortItemOrdering"
    nullOrdering: "SortItemNullOrdering"

    def __init__(
        self,
        sortKey: "Expression",
        ordering: "SortItemOrdering",
        nullOrdering: "SortItemNullOrdering"
    ) -> None:
        super().__init__(_type="sortItem")
        self.sortKey = sortKey
        self.ordering = ordering
        self.nullOrdering = nullOrdering


class ComparisonExpression(Expression):
    left: "Expression"
    right: "Expression"
    operator: "ComparisonOperator"

    def __init__(
        self,
        left: "Expression",
        right: "Expression",
        operator: "ComparisonOperator"
    ) -> None:
        super().__init__(_type="comparisonExpression")
        self.left = left
        self.right = right
        self.operator = operator


class LogicalBinaryExpression(Expression):
    type_: "LogicalBinaryType"
    left: "Expression"
    right: "Expression"

    def __init__(
        self,
        type_: "LogicalBinaryType",
        left: "Expression",
        right: "Expression"
    ) -> None:
        super().__init__(_type="logicalBinaryExpression")
        self.type_ = type_
        self.left = left
        self.right = right


class NotExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression"
    ) -> None:
        super().__init__(_type="notExpression")
        self.value = value


class ArithmeticExpression(Expression):
    type_: "ArithmeticType"
    left: "Expression"
    right: "Expression"

    def __init__(
        self,
        type_: "ArithmeticType",
        left: "Expression",
        right: "Expression"
    ) -> None:
        super().__init__(_type="arithmeticExpression")
        self.type_ = type_
        self.left = left
        self.right = right


class NegativeExpression(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression"
    ) -> None:
        super().__init__(_type="negativeExpression")
        self.value = value


class IsNullPredicate(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression"
    ) -> None:
        super().__init__(_type="isNullPredicate")
        self.value = value


class IsNotNullPredicate(Expression):
    value: "Expression"

    def __init__(
        self,
        value: "Expression"
    ) -> None:
        super().__init__(_type="isNotNullPredicate")
        self.value = value


class CurrentTime(Expression):
    type_: "CurrentTimeType"
    precision: "PyLegendOptional[int]"

    def __init__(
        self,
        type_: "CurrentTimeType",
        precision: "PyLegendOptional[int]"
    ) -> None:
        super().__init__(_type="currentTime")
        self.type_ = type_
        self.precision = precision


class FunctionCall(Expression):
    name: "QualifiedName"
    distinct: "bool"
    arguments: "PyLegendList[Expression]"
    filter_: "PyLegendOptional[Expression]"
    window: "PyLegendOptional[Window]"

    def __init__(
        self,
        name: "QualifiedName",
        distinct: "bool",
        arguments: "PyLegendList[Expression]",
        filter_: "PyLegendOptional[Expression]",
        window: "PyLegendOptional[Window]"
    ) -> None:
        super().__init__(_type="functionCall")
        self.name = name
        self.distinct = distinct
        self.arguments = arguments
        self.filter_ = filter_
        self.window = window


class SimpleCaseExpression(Expression):
    operand: "Expression"
    whenClauses: "PyLegendList[WhenClause]"
    defaultValue: "PyLegendOptional[Expression]"

    def __init__(
        self,
        operand: "Expression",
        whenClauses: "PyLegendList[WhenClause]",
        defaultValue: "PyLegendOptional[Expression]"
    ) -> None:
        super().__init__(_type="simpleCaseExpression")
        self.operand = operand
        self.whenClauses = whenClauses
        self.defaultValue = defaultValue


class SearchedCaseExpression(Expression):
    whenClauses: "PyLegendList[WhenClause]"
    defaultValue: "PyLegendOptional[Expression]"

    def __init__(
        self,
        whenClauses: "PyLegendList[WhenClause]",
        defaultValue: "PyLegendOptional[Expression]"
    ) -> None:
        super().__init__(_type="searchedCaseExpression")
        self.whenClauses = whenClauses
        self.defaultValue = defaultValue


class WhenClause(Expression):
    operand: "Expression"
    result: "Expression"

    def __init__(
        self,
        operand: "Expression",
        result: "Expression"
    ) -> None:
        super().__init__(_type="whenClause")
        self.operand = operand
        self.result = result


class Extract(Expression):
    expression: "Expression"
    field: "ExtractField"

    def __init__(
        self,
        expression: "Expression",
        field: "ExtractField"
    ) -> None:
        super().__init__(_type="extract")
        self.expression = expression
        self.field = field


class Join(Relation):
    type_: "JoinType"
    left: "Relation"
    right: "Relation"
    criteria: "PyLegendOptional[JoinCriteria]"

    def __init__(
        self,
        type_: "JoinType",
        left: "Relation",
        right: "Relation",
        criteria: "PyLegendOptional[JoinCriteria]"
    ) -> None:
        super().__init__(_type="join")
        self.type_ = type_
        self.left = left
        self.right = right
        self.criteria = criteria


class JoinCriteria:
    _type: str

    def __init__(
        self,
        _type: str
    ) -> None:
        self._type = _type


class JoinOn(JoinCriteria):
    expression: "Expression"

    def __init__(
        self,
        expression: "Expression"
    ) -> None:
        super().__init__(_type="joinOn")
        self.expression = expression


class JoinUsing(JoinCriteria):
    columns: "PyLegendList[str]"

    def __init__(
        self,
        columns: "PyLegendList[str]"
    ) -> None:
        super().__init__(_type="joinUsing")
        self.columns = columns


class Cast(Expression):
    expression: "Expression"
    type_: "ColumnType"

    def __init__(
        self,
        expression: "Expression",
        type_: "ColumnType"
    ) -> None:
        super().__init__(_type="cast")
        self.expression = expression
        self.type_ = type_


class ColumnType(Expression):
    name: "str"
    parameters: "PyLegendList[int]"

    def __init__(
        self,
        name: "str",
        parameters: "PyLegendList[int]"
    ) -> None:
        super().__init__(_type="columnType")
        self.name = name
        self.parameters = parameters


class QualifiedName:
    parts: "PyLegendList[str]"

    def __init__(
        self,
        parts: "PyLegendList[str]"
    ) -> None:
        self.parts = parts


class QualifiedNameReference(Expression):
    name: "QualifiedName"

    def __init__(
        self,
        name: "QualifiedName"
    ) -> None:
        super().__init__(_type="qualifiedNameReference")
        self.name = name


class InListExpression(Expression):
    values: "PyLegendList[Expression]"

    def __init__(
        self,
        values: "PyLegendList[Expression]"
    ) -> None:
        super().__init__(_type="inListExpression")
        self.values = values


class InPredicate(Expression):
    value: "Expression"
    valueList: "Expression"

    def __init__(
        self,
        value: "Expression",
        valueList: "Expression"
    ) -> None:
        super().__init__(_type="inPredicate")
        self.value = value
        self.valueList = valueList


class Window(Statement):
    windowRef: "PyLegendOptional[str]"
    partitions: "PyLegendList[Expression]"
    orderBy: "PyLegendList[SortItem]"
    windowFrame: "PyLegendOptional[WindowFrame]"

    def __init__(
        self,
        windowRef: "PyLegendOptional[str]",
        partitions: "PyLegendList[Expression]",
        orderBy: "PyLegendList[SortItem]",
        windowFrame: "PyLegendOptional[WindowFrame]"
    ) -> None:
        super().__init__(_type="window")
        self.windowRef = windowRef
        self.partitions = partitions
        self.orderBy = orderBy
        self.windowFrame = windowFrame


class WindowFrame(Node):
    mode: "WindowFrameMode"
    start: "FrameBound"
    end: "PyLegendOptional[FrameBound]"

    def __init__(
        self,
        mode: "WindowFrameMode",
        start: "FrameBound",
        end: "PyLegendOptional[FrameBound]"
    ) -> None:
        super().__init__(_type="windowFrame")
        self.mode = mode
        self.start = start
        self.end = end


class FrameBound(Node):
    type_: "FrameBoundType"
    value: "PyLegendOptional[Expression]"
    duration_unit: "PyLegendOptional[StringLiteral]"

    def __init__(
        self,
        type_: "FrameBoundType",
        value: "PyLegendOptional[Expression]",
        duration_unit: "PyLegendOptional[StringLiteral]" = None
    ) -> None:
        super().__init__(_type="frameBound")
        self.type_ = type_
        self.value = value
        self.duration_unit = duration_unit



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendList,
    PyLegendSequence,
    PyLegendUnion,
    PyLegendOptional,
    PyLegendTuple,
)
from pylegend.core.language import (
    PyLegendBoolean,
    PyLegendBooleanLiteralExpression,
    PyLegendPrimitive,
)
from pylegend.core.language.pandas_api.pandas_api_tds_row import PandasApiTdsRow
from pylegend.core.language.shared.helpers import generate_pure_lambda
from pylegend.core.language.shared.literal_expressions import convert_literal_to_literal_expression
from pylegend.core.sql.metamodel import (
    QuerySpecification,
    Select,
    SelectItem,
    SingleColumn,
    AliasedRelation,
    TableSubquery,
    Query,
    Join,
    JoinType,
    JoinOn,
    QualifiedNameReference,
    QualifiedName,
)
from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import PandasApiAppliedFunction
from pylegend.core.tds.pandas_api.frames.pandas_api_base_tds_frame import PandasApiBaseTdsFrame
from pylegend.core.tds.pandas_api.frames.pandas_api_tds_frame import PandasApiTdsFrame
from pylegend.core.tds.sql_query_helpers import copy_query, create_sub_query, extract_columns_for_subquery
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToSqlConfig, FrameToPureConfig

__all__: PyLegendSequence[str] = [
    "PandasApiMergeFunction"
]


class PandasApiMergeFunction(PandasApiAppliedFunction):
    __base_frame: PandasApiBaseTdsFrame
    __other_frame: PandasApiBaseTdsFrame
    __how: PyLegendOptional[str]
    __on: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str]]]
    __left_on: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str]]]
    __right_on: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str]]]
    __left_index: PyLegendOptional[bool]
    __right_index: PyLegendOptional[bool]
    __sort: PyLegendOptional[bool]
    __suffixes: PyLegendOptional[
        PyLegendUnion[
            PyLegendTuple[PyLegendUnion[str, None], PyLegendUnion[str, None]],
            PyLegendList[PyLegendUnion[str, None]],
        ]
    ]
    __indicator: PyLegendOptional[PyLegendUnion[bool, str]]
    __validate: PyLegendOptional[str]

    @classmethod
    def name(cls) -> str:
        return "merge"  # pragma: no cover

    def __init__(
            self,
            base_frame: PandasApiBaseTdsFrame,
            other_frame: PandasApiBaseTdsFrame,
            how: PyLegendOptional[str],
            on: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str]]],
            left_on: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str]]],
            right_on: PyLegendOptional[PyLegendUnion[str, PyLegendSequence[str]]],
            left_index: PyLegendOptional[bool],
            right_index: PyLegendOptional[bool],
            sort: PyLegendOptional[bool],
            suffixes: PyLegendOptional[
                PyLegendUnion[
                    PyLegendTuple[PyLegendUnion[str, None], PyLegendUnion[str, None]],
                    PyLegendList[PyLegendUnion[str, None]],
                ]
            ],
            indicator: PyLegendOptional[PyLegendUnion[bool, str]],
            validate: PyLegendOptional[str]
    ) -> None:
        self.__base_frame = base_frame
        self.__other_frame = other_frame
        self.__on = on
        self.__left_on = left_on
        self.__right_on = right_on
        self.__left_index = left_index
        self.__right_index = right_index
        self.__sort = sort
        self.__how = how
        self.__suffixes = suffixes
        self.__indicator = indicator
        self.__validate = validate
        self.__sortkeys = []  # type: PyLegendList[str]

    def get_sort_keys(self) -> PyLegendList[str]:
        return self.__sortkeys

    # Key resolution helpers
    def __normalize_keys(
            self,
            candidate: PyLegendUnion[str, PyLegendSequence[str], None]
    ) -> PyLegendList[str]:
        if candidate is None:
            return []
        return [candidate] if isinstance(candidate, str) else list(candidate)

    def __derive_key_pairs(self) -> PyLegendList[PyLegendTuple[str, str]]:

        if self.__how.lower() == "cross":  # type: ignore
            return []

        left_cols = [c.get_name() for c in self.__base_frame.columns()]
        right_cols = [c.get_name() for c in self.__other_frame.columns()]

        if self.__on is not None and (self.__left_on is not None or self.__right_on is not None):
            raise ValueError('Can only pass argument "on" OR "left_on" and "right_on", not a combination of both.')
        if self.__on is not None:
            on_keys = self.__normalize_keys(self.__on)
            for k in on_keys:
                if k not in left_cols or k not in right_cols:
                    raise KeyError(f"'{k}' not found")
            return [(k, k) for k in on_keys]

        left_keys = self.__normalize_keys(self.__left_on)
        right_keys = self.__normalize_keys(self.__right_on)

        if left_keys or right_keys:
            if len(left_keys) != len(right_keys):
                print("came here")
                raise ValueError("len(right_on) must equal len(left_on)")
            for lk in left_keys:
                if lk not in left_cols:
                    raise KeyError(f"'{lk}' not found")
            for rk in right_keys:
                if rk not in right_cols:
                    raise KeyError(f"'{rk}' not found")

            return list(zip(left_keys, right_keys))

        # Infer intersection by default
        inferred = [c for c in left_cols if c in right_cols]
        return [(k, k) for k in inferred]

    def __normalize_suffixes(self) -> None:
        # Convert None to empty string and coerce to list[str]
        left = self.__suffixes[0] or ""  # type: ignore
        right = self.__suffixes[1] or ""  # type: ignore

        self.__suffixes = [left, right]

    # Internal auto join condition builder (returns PyLegendBoolean expression)
    def __build_condition(self) -> PyLegendBoolean:
        key_pairs = self.__derive_key_pairs()
        left_row = PandasApiTdsRow.from_tds_frame("left", self.__base_frame)
        right_row = PandasApiTdsRow.from_tds_frame("right", self.__other_frame)

        expr = None
        for left_key, right_key in key_pairs:
            part = (left_row[left_key] == right_row[right_key])
            expr = part if expr is None else (expr & part)
        return expr  # type: ignore

    def __join_type(self) -> JoinType:
        how_lower = self.__how.lower()  # type: ignore
        if how_lower == "inner":
            return JoinType.INNER
        if how_lower == "left":
            return JoinType.LEFT
        if how_lower == "right":
            return JoinType.RIGHT
        if how_lower == "outer":
            return JoinType.FULL
        if how_lower == "cross":
            return JoinType.CROSS
        raise ValueError("do not recognize join method " + self.__how)  # type: ignore

    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        db_extension = config.sql_to_string_generator().get_db_extension()
        left_query = copy_query(self.__base_frame.to_sql_query_object(config))
        right_query = copy_query(self.__other_frame.to_sql_query_object(config))

        join_criteria = None
        join_type = self.__join_type()
        if join_type != JoinType.CROSS:
            join_condition_expr = self.__build_condition()
            if isinstance(join_condition_expr, bool):
                join_condition_expr = PyLegendBoolean(PyLegendBooleanLiteralExpression(join_condition_expr))  # pragma: no cover  # noqa: E501
            join_sql_expr = join_condition_expr.to_sql_expression(
                {
                    "left": create_sub_query(left_query, config, "left"),
                    "right": create_sub_query(right_query, config, "right"),
                },
                config
            )
            join_criteria = JoinOn(expression=join_sql_expr)

        left_alias = db_extension.quote_identifier("left")
        right_alias = db_extension.quote_identifier("right")

        left_original = {c.get_name(): c for c in self.__base_frame.columns()}
        right_original = {c.get_name(): c for c in self.__other_frame.columns()}
        key_pairs = self.__derive_key_pairs()
        same_name_keys = {left_key for left_key, right_key in key_pairs if left_key == right_key}

        select_items: PyLegendList[SelectItem] = []

        left_cols_set = set(left_original.keys())
        right_cols_set = set(right_original.keys())
        overlapping = (left_cols_set & right_cols_set) - same_name_keys

        # Left select items
        for c in self.__base_frame.columns():
            orig = c.get_name()
            out_name = orig + self.__suffixes[0] if orig in overlapping else orig  # type: ignore
            q_out = db_extension.quote_identifier(out_name)
            q_in = db_extension.quote_identifier(orig)
            select_items.append(
                SingleColumn(q_out, QualifiedNameReference(QualifiedName(parts=[left_alias, q_in])))
            )

        # Right select items
        for c in self.__other_frame.columns():
            orig = c.get_name()
            if orig in same_name_keys:
                continue

            out_name = orig + self.__suffixes[1] if orig in overlapping else orig  # type: ignore
            q_out = db_extension.quote_identifier(out_name)
            q_in = db_extension.quote_identifier(orig)
            select_items.append(
                SingleColumn(q_out, QualifiedNameReference(QualifiedName(parts=[right_alias, q_in])))
            )

        join_spec = QuerySpecification(
            select=Select(selectItems=select_items, distinct=False),
            from_=[
                Join(
                    type_=self.__join_type(),
                    left=AliasedRelation(
                        relation=TableSubquery(Query(queryBody=left_query, limit=None, offset=None, orderBy=[])),
                        alias=left_alias,
                        columnNames=extract_columns_for_subquery(left_query)
                    ),
                    right=AliasedRelation(
                        relation=TableSubquery(Query(queryBody=right_query, limit=None, offset=None, orderBy=[])),
                        alias=right_alias,
                        columnNames=extract_columns_for_subquery(right_query)
                    ),
                    criteria=join_criteria
                )
            ],
            where=None,
            groupBy=[],
            having=None,
            orderBy=[],
            limit=None,
            offset=None
        )

        return create_sub_query(join_spec, config, "root")

    def to_pure(self, config: FrameToPureConfig) -> str:
        how_lower = self.__how.lower()  # type: ignore
        if how_lower == "inner":
            join_kind = "INNER"
        elif how_lower == "left":
            join_kind = "LEFT"
        elif how_lower == "right":
            join_kind = "RIGHT"
        elif how_lower == "outer":
            join_kind = "FULL"
        elif how_lower == "cross":
            join_kind = "INNER"

        # Resolve key pairs
        key_pairs = self.__derive_key_pairs()

        left_cols = [c.get_name() for c in self.__base_frame.columns()]
        right_cols = [c.get_name() for c in self.__other_frame.columns()]

        # Suffix handling for overlapping non-key columns
        s = list(self.__suffixes)  # type: ignore
        left_suf, right_suf = s[0], s[1]

        left_col_set = set(left_cols)
        right_col_set = set(right_cols)
        overlapping = left_col_set & right_col_set

        # Overlapping join keys (same name) need temporary rename on right to allow join
        identical_key_names = {
            lk for (lk, rk) in key_pairs if lk == rk
        }

        left_rename_map = {}
        right_rename_map = {}

        # Non-key overlapping columns get suffixes
        for col in overlapping:
            if col in identical_key_names:
                continue

            left_rename_map[col] = col + left_suf  # type: ignore
            right_rename_map[col] = col + right_suf  # type: ignore

        # Temporary rename for identical key names on right
        temp_right_key_map = {}
        for k in identical_key_names:
            temp_name = k + "__right_key_tmp"
            temp_right_key_map[k] = temp_name

        # Rename
        left_frame = (self.__base_frame.rename(columns=left_rename_map, errors="raise")
                      if left_rename_map else self.__base_frame)

        right_map = {**right_rename_map, **temp_right_key_map} if (right_rename_map or temp_right_key_map) else None
        right_frame = (self.__other_frame.rename(columns=right_map, errors="raise")
                       if right_map else self.__other_frame)

        # Build join condition expression
        if how_lower != "cross":
            left_row = PandasApiTdsRow.from_tds_frame("l", left_frame)
            right_row = PandasApiTdsRow.from_tds_frame("r", right_frame)

            expr = None
            for l_key, r_key in key_pairs:
                l_eff = left_rename_map.get(l_key, l_key)
                r_eff = right_map.get(r_key, r_key)  # type: ignore
                part = (left_row[l_eff] == right_row[r_eff])
                expr = part if expr is None else (expr & part)

            if not isinstance(expr, PyLegendPrimitive):
                expr = convert_literal_to_literal_expression(expr)  # type: ignore  # pragma: no cover
            cond_str = expr.to_pure_expression(config.push_indent(2))  # type: ignore
        else:
            cond_str = "1==1"

        left_pure = left_frame.to_pure(config)  # type: ignore
        right_pure = right_frame.to_pure(config.push_indent(2))  # type: ignore

        join_expr = (
            f"{left_pure}{config.separator(1)}"
            f"->join({config.separator(2)}"
            f"{right_pure},{config.separator(2, True)}"
            f"JoinKind.{join_kind},{config.separator(2, True)}"
            f"{generate_pure_lambda('l, r', cond_str)}{config.separator(1)})"
        )

        # Only project if temporary right key renames exist
        if temp_right_key_map:
            final_cols = []

            for c in left_cols:
                if (c in overlapping and c not in identical_key_names):
                    final_cols.append(c + left_suf)  # type: ignore
                else:
                    final_cols.append(c)

            for c in right_cols:
                if c in temp_right_key_map:
                    continue
                if (c in overlapping and c not in identical_key_names):
                    final_cols.append(c + right_suf)  # type: ignore
                else:
                    final_cols.append(c)

            project_items = [f"{col}:x|$x.{col}" for col in final_cols]
            project_body = ", ".join(project_items)
            join_expr = (
                f"{join_expr}{config.separator(1)}"
                f"->project({config.separator(2)}~[{project_body}]{config.separator(1)})"
            )

        return join_expr

    def base_frame(self) -> PandasApiBaseTdsFrame:
        return self.__base_frame

    def tds_frame_parameters(self) -> PyLegendList["PandasApiBaseTdsFrame"]:
        return [self.__other_frame]

    def calculate_columns(self) -> PyLegendSequence["TdsColumn"]:
        key_pairs = self.__derive_key_pairs()
        left_keys_same_name = {left_key for left_key, right_key in key_pairs if left_key == right_key}
        left_cols = [c.get_name() for c in self.__base_frame.columns()]
        right_cols = [c.get_name() for c in self.__other_frame.columns()]

        overlapping = (set(left_cols) & set(right_cols)) - left_keys_same_name

        # Build left columns (apply suffix to overlapping non-key)
        result_cols: PyLegendSequence["TdsColumn"] = []
        for c in self.__base_frame.columns():
            name = c.get_name()
            if name in overlapping:
                result_cols.append(c.copy_with_changed_name(name + self.__suffixes[0]))  # type: ignore
            else:
                result_cols.append(c.copy())  # type: ignore

        # Build right columns (skip same-name keys; apply suffix to overlapping non-key)
        for c in self.__other_frame.columns():
            name = c.get_name()
            if any(name == left_key and left_key == right_key for left_key, right_key in key_pairs):
                continue

            if name in overlapping:
                result_cols.append(c.copy_with_changed_name(name + self.__suffixes[1]))  # type: ignore
            else:
                result_cols.append(c.copy())  # type: ignore

        # Validate no duplicates
        names = [c.get_name() for c in result_cols]
        if len(names) != len(set(names)):
            raise ValueError("Resulting merged columns contain duplicates after suffix application")

        if self.__sort:
            for lk, rk in key_pairs:
                left_out = (lk + self.__suffixes[0]) if (lk in overlapping) else lk  # type: ignore
                self.__sortkeys.append(left_out)

                if rk != lk:
                    right_out = (rk + self.__suffixes[1]) if (rk in overlapping) else rk  # type: ignore
                    # right identical-name keys are skipped
                    self.__sortkeys.append(right_out)

        return result_cols

    def validate(self) -> bool:
        # Frame type validation
        if not isinstance(self.__other_frame, PandasApiTdsFrame):
            raise TypeError(f"Can only merge TdsFrame objects, a {type(self.__other_frame)} was passed")

        # Same frame not supported
        if self.__base_frame is self.__other_frame:
            raise NotImplementedError("Merging the same TdsFrame is not supported yet")

        # how
        if not isinstance(self.__how, str):
            raise TypeError(f"'how' must be str, got {type(self.__how)}")

        if self.__how.lower() == "cross":
            if any(v is not None for v in (self.__on, self.__left_on, self.__right_on)):
                raise ValueError("Can not pass on, right_on, left_on for how='cross'")

        # key parameters: on / left_on / right_on
        def _validate_keys_param(param_name: str, value: PyLegendUnion[str, PyLegendSequence[str], None]) -> None:
            if value is None:
                return
            if isinstance(value, str):
                return
            if isinstance(value, (list, tuple)):
                if not all(isinstance(v, str) for v in value):
                    raise TypeError(f"'{param_name}' must contain only str elements")
                return
            raise TypeError(
                f"Passing '{param_name}' as a {type(value)} is not supported. "
                f"Provide '{param_name}' as a tuple instead."
            )

        _validate_keys_param("on", self.__on)
        _validate_keys_param("left_on", self.__left_on)
        _validate_keys_param("right_on", self.__right_on)

        # Suffix validation
        if not isinstance(self.__suffixes, (tuple, list)):
            raise TypeError(
                f"Passing 'suffixes' as {type(self.__suffixes)}, is not supported. "
                "Provide 'suffixes' as a tuple instead."
            )
        for s in self.__suffixes:
            if s is not None and not isinstance(s, str):
                raise TypeError("'suffixes' elements must be str or None")
        if len(self.__suffixes) != 2:
            raise ValueError("too many values to unpack (expected 2)")

        # Sort
        if self.__sort is not None and not isinstance(self.__sort, (bool, PyLegendBoolean)):
            raise TypeError(f"Sort parameter must be bool, got {type(self.__sort)}")

        # Unsupported parameters
        if self.__left_index or self.__right_index:
            raise NotImplementedError("Merging on index is not supported yet in PandasApi merge function")

        if self.__indicator:
            raise NotImplementedError("Indicator parameter is not supported yet in PandasApi merge function")

        if self.__validate:
            raise NotImplementedError("Validate parameter is not supported yet in PandasApi merge function")

        self.__join_type()  # runs how validation
        self.__normalize_suffixes()  # runs suffixes validation

        key_pairs = self.__derive_key_pairs()  # runs key validations
        if not key_pairs and self.__how.lower() != "cross":
            raise ValueError("No merge keys resolved. Specify 'on' or 'left_on'/'right_on', or ensure common columns.")

        return True



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from pylegend._typing import (
    PyLegendUnion,
    PyLegendOptional,
    PyLegendSequence,
    PyLegendList,
    PyLegendTuple
)
from pylegend.core.language import (
    PyLegendInteger,
)
from pylegend.core.language.shared.helpers import escape_column_name
from pylegend.core.sql.metamodel import (
    QuerySpecification,
    SingleColumn,
    SelectItem
)
from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
    PandasApiAppliedFunction,
)
from pylegend.core.tds.pandas_api.frames.pandas_api_base_tds_frame import (
    PandasApiBaseTdsFrame,
)
from pylegend.core.tds.sql_query_helpers import copy_query, create_sub_query
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.tds.tds_frame import FrameToSqlConfig, FrameToPureConfig

__all__: PyLegendSequence[str] = ["PandasApiFilterFunction"]


class PandasApiFilterFunction(PandasApiAppliedFunction):
    __base_frame: PandasApiBaseTdsFrame
    __items: PyLegendOptional[PyLegendList[str]]
    __like: PyLegendOptional[str]
    __regex: PyLegendOptional[str]
    __axis: PyLegendOptional[PyLegendUnion[str, int, PyLegendInteger]]

    @classmethod
    def name(cls) -> str:
        return "filter"  # pragma: no cover

    def __init__(
            self,
            base_frame: PandasApiBaseTdsFrame,
            items: PyLegendOptional[PyLegendList[str]],
            like: PyLegendOptional[str],
            regex: PyLegendOptional[str],
            axis: PyLegendOptional[PyLegendUnion[str, int, PyLegendInteger]],
    ) -> None:
        self.__base_frame = base_frame
        self.__items = items
        self.__like = like
        self.__regex = regex
        self.__axis = 1 if axis is None else axis

    def __get_desired_columns(
            self, col_names: PyLegendSequence[str]
    ) -> PyLegendSequence[str]:
        if self.__items is not None:
            return self.__items
        elif self.__like is not None:
            return [col for col in col_names if self.__like in col]
        elif self.__regex is not None:
            regex_pattern = re.compile(self.__regex)
            return [col for col in col_names if regex_pattern.search(col)]

        return []  # pragma: no cover

    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        base_query = self.__base_frame.to_sql_query_object(config)
        db_extension = config.sql_to_string_generator().get_db_extension()
        columns_to_retain = [db_extension.quote_identifier(x) for x in
                             self.__get_desired_columns([c.get_name() for c in self.__base_frame.columns()])]

        sub_query_required = (
                len(base_query.groupBy) > 0 or
                len(base_query.orderBy) > 0 or
                base_query.having is not None or
                base_query.select.distinct
        )

        if sub_query_required:
            new_query = create_sub_query(base_query, config, "root", columns_to_retain=columns_to_retain)
            return new_query
        else:
            new_cols_with_index: PyLegendList[PyLegendTuple[int, SelectItem]] = []
            for col in base_query.select.selectItems:
                if not isinstance(col, SingleColumn):
                    raise ValueError(
                        "Select operation not supported for queries with columns other than SingleColumn"
                    )  # pragma: no cover
                if col.alias is None:
                    raise ValueError(
                        "Select operation not supported for queries with SingleColumns with missing alias"
                    )  # pragma: no cover
                if col.alias in columns_to_retain:
                    new_cols_with_index.append((columns_to_retain.index(col.alias), col))

            new_select_items = [y[1] for y in sorted(new_cols_with_index, key=lambda x: x[0])]
            new_query = copy_query(base_query)
            new_query.select.selectItems = new_select_items
            return new_query

    def to_pure(self, config: FrameToPureConfig) -> str:
        col_names = [c.get_name() for c in self.__base_frame.columns()]
        desired_columns = self.__get_desired_columns(col_names)
        escaped_columns = [escape_column_name(col_name) for col_name in desired_columns]
        return (
            f"{self.__base_frame.to_pure(config)}{config.separator(1)}"
            f"->select(~[{', '.join(escaped_columns)}])"
        )

    def base_frame(self) -> PandasApiBaseTdsFrame:
        return self.__base_frame

    def tds_frame_parameters(self) -> PyLegendList["PandasApiBaseTdsFrame"]:
        return []

    def calculate_columns(self) -> PyLegendSequence["TdsColumn"]:
        base_cols = [c.copy() for c in self.__base_frame.columns()]
        desired_col_names = self.__get_desired_columns([c.get_name() for c in base_cols])
        base_col_map = {c.get_name(): c for c in base_cols}
        return [
            base_col_map[name].copy()
            for name in desired_col_names
            if name in base_col_map
        ]

    def validate(self) -> bool:
        mutual_exclusion = sum(
            [
                self.__items is not None,
                self.__like is not None,
                self.__regex is not None,
            ]
        )
        if mutual_exclusion > 1:
            raise TypeError(
                "Keyword arguments `items`, `like`, or `regex` are mutually exclusive"
            )
        if mutual_exclusion == 0:
            raise TypeError("Must pass either `items`, `like`, or `regex`")

        base_cols = [c.get_name() for c in self.__base_frame.columns()]
        if self.__items is not None:
            if not isinstance(self.__items, (list, PyLegendList)):
                raise TypeError(
                    f"Index(...) must be called with a collection, got '{self.__items}'"
                )
            invalid_cols = [item for item in self.__items if item not in base_cols]
            if invalid_cols:
                raise ValueError(
                    f"Columns {invalid_cols} in `filter` items list do not exist. Available: {base_cols}"
                )

        if self.__like is not None:
            if not isinstance(self.__like, str):
                raise TypeError(f"'like' must be a string, got {type(self.__like)}")
            if not any(self.__like in col for col in base_cols):
                raise ValueError(
                    f"No columns match the pattern '{self.__like}'. Available: {base_cols}"
                )

        if self.__regex is not None:
            if not isinstance(self.__regex, str):
                raise TypeError(f"'regex' must be a string, got {type(self.__regex)}")
            try:
                regex_pattern = re.compile(self.__regex)
            except re.error as e:
                raise ValueError(f"Invalid regex pattern '{self.__regex}': {e}")
            if not any(regex_pattern.search(col) for col in base_cols):
                raise ValueError(
                    f"No columns match the regex '{self.__regex}'. Available: {base_cols}"
                )

        if not isinstance(
                self.__axis, (str, int, PyLegendInteger)
        ) or self.__axis not in [1, "columns"]:
            raise ValueError(
                f"Unsupported axis value: {self.__axis}. Expected 1 or 'columns'"
            )

        return True



# Copyright 2025 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pylegend._typing import (
    PyLegendUnion,
    PyLegendOptional,
    PyLegendCallable,
    PyLegendList,
    PyLegendSequence,
)
from pylegend.core.language.shared.tds_row import AbstractTdsRow
from pylegend.core.sql.metamodel import (
    QuerySpecification,
    SortItemOrdering,
    SortItem,
    SortItemNullOrdering, SingleColumn, Expression,
)
from pylegend.core.tds.pandas_api.frames.pandas_api_applied_function_tds_frame import (
    PandasApiAppliedFunction,
)
from pylegend.core.tds.pandas_api.frames.pandas_api_base_tds_frame import (
    PandasApiBaseTdsFrame,
)
from pylegend.core.tds.sql_query_helpers import create_sub_query, copy_query
from pylegend.core.tds.tds_frame import FrameToSqlConfig, FrameToPureConfig
from pylegend.core.tds.tds_column import TdsColumn
from pylegend.core.language.shared.helpers import escape_column_name


class SortValuesFunction(PandasApiAppliedFunction):
    __base_frame: PandasApiBaseTdsFrame
    __by: PyLegendList[str]
    __axis: PyLegendUnion[str, int]
    __ascending: PyLegendList[bool]
    __inplace: bool
    __kind: PyLegendOptional[str]
    __na_position: str
    __ignore_index: bool
    key: PyLegendOptional[PyLegendCallable[[AbstractTdsRow], AbstractTdsRow]] = None

    @classmethod
    def name(cls) -> str:
        return "sort_values"  # pragma: no cover

    def __init__(
            self,
            base_frame: PandasApiBaseTdsFrame,
            by: PyLegendUnion[str, PyLegendList[str]],
            axis: PyLegendUnion[str, int],
            ascending: PyLegendUnion[bool, PyLegendList[bool]],
            inplace: bool,
            kind: PyLegendOptional[str],
            na_position: str,
            ignore_index: bool,
            key: PyLegendOptional[PyLegendCallable[[AbstractTdsRow], AbstractTdsRow]] = None
    ) -> None:
        self.__base_frame = base_frame
        self.__by_input = by
        self.__axis = axis
        self.__ascending_input = ascending
        self.__inplace = inplace
        self.__kind = kind
        self.__na_position = na_position
        self.__ignore_index = ignore_index
        self.__key = key

    def to_sql(self, config: FrameToSqlConfig) -> QuerySpecification:
        base_query: QuerySpecification = self.__base_frame.to_sql_query_object(config)
        should_create_sub_query = (base_query.offset is not None) or (base_query.limit is not None)
        new_query = (
            create_sub_query(base_query, config, "root") if should_create_sub_query
            else copy_query(base_query)
        )
        new_query.orderBy = [
            SortItem(
                sortKey=self.get_expression_from_column_name(new_query, column_name, config),
                ordering=(
                    SortItemOrdering.ASCENDING if ascending
                    else SortItemOrdering.DESCENDING
                ),
                nullOrdering=SortItemNullOrdering.UNDEFINED,
            )
            for column_name, ascending in zip(self.__by, self.__ascending)
        ]
        return new_query

    def get_expression_from_column_name(self, query: QuerySpecification, column_name: str,
                                        config: FrameToSqlConfig) -> Expression:
        db_extension = config.sql_to_string_generator().get_db_extension()
        filtered = [
            s for s in query.select.selectItems
            if (isinstance(s, SingleColumn) and
                s.alias == db_extension.quote_identifier(column_name))
        ]
        if len(filtered) == 0:
            raise RuntimeError("Cannot find column: " + column_name)  # pragma: no cover
        return filtered[0].expression

    def to_pure(self, config: FrameToPureConfig) -> str:
        escaped_columns = []
        for col_name in self.__by:
            escaped_columns.append(escape_column_name(col_name))
        sort_items = [
            f"~{column_name}->ascending()" if ascending else f"~{column_name}->descending()"
            for column_name, ascending in zip(escaped_columns, self.__ascending)
        ]
        return (
                f"{self.__base_frame.to_pure(config)}{config.separator(1)}"
                + f"->sort([{', '.join(sort_items)}])"
        )

    def base_frame(self) -> PandasApiBaseTdsFrame:
        return self.__base_frame

    def tds_frame_parameters(self) -> PyLegendList["PandasApiBaseTdsFrame"]:
        return []

    def calculate_columns(self) -> PyLegendSequence["TdsColumn"]:
        return [c.copy() for c in self.__base_frame.columns()]

    def validate(self) -> bool:
        if self.__axis not in [0, "index"]:
            raise ValueError(
                "Axis parameter of sort_values function must be 0 or 'index'"
            )

        if self.__inplace is not False:
            raise ValueError("Inplace parameter of sort_values function must be False")

        if self.__kind is not None:
            raise NotImplementedError(
                "Kind parameter of sort_values function is not supported"
            )

        if self.__ignore_index is not True:
            raise ValueError(
                "Ignore_index parameter of sort_values function must be True"
            )

        if self.__key is not None:
            raise NotImplementedError(
                "Key parameter of sort_values function is not supported"
            )

        base_frame_columns = [
            column.get_name() for column in self.__base_frame.columns()
        ]

        self.__by = self._build_by_list()
        self.__ascending = self._build_ascending_list()

        if len(self.__by) != len(self.__ascending):
            raise ValueError(
                "The number of columns in 'by' must equal the number of values in 'ascending' for sort_values function."
            )

        for column in self.__by:
            if column not in base_frame_columns:
                raise ValueError(
                    f"Column - '{column}' in sort_values columns list doesn't exist in the current frame. "
                    f"Current frame columns: {base_frame_columns}"
                )

        return True

    def _build_by_list(self) -> PyLegendList[str]:
        if isinstance(self.__by_input, str):
            return [self.__by_input]
        else:
            return self.__by_input

    def _build_ascending_list(self) -> PyLegendList[bool]:
        if self.__ascending_input is True:
            return [True for _ in self.__by]
        elif self.__ascending_input is False:
            return [False for _ in self.__by]
        else:
            return self.__ascending_input
